/**
 * @file lock.cpp
 *
 * @ingroup iopccommon
 * @author Petr Cermak
 */

#include <pthread.h>
#include <errno.h>

#include "lock.h"
#include "defs.h"
#include "trace.h"

namespace iopc {
Lock::Lock()
	throw (IopcNoMemoryException, IopcRsrcLackException)
{
	switch(pthread_mutexattr_init(&mutexattr)) {
	case 0:
	break;
	case ENOMEM:
		throw IopcNoMemoryException();
	default:
		IOPC_UNREACHABLE;
	break;
	}

	if (pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_RECURSIVE) != 0) {
		IOPC_UNREACHABLE;
	}

	switch (pthread_mutex_init(&mutex, &mutexattr)) {
	case 0:
	break;
	case ENOMEM:
		throw IopcNoMemoryException();
	break;
	case EAGAIN:
	case EPERM:
		throw IopcRsrcLackException();
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

Lock::~Lock()
	throw ()
{
	int res = pthread_mutex_destroy(&mutex);
	if (res != 0) {
		IOPC_UNREACHABLE;
	}
	res = pthread_mutexattr_destroy(&mutexattr);
	if (res != 0) {
		IOPC_UNREACHABLE;
	}
}

void Lock::lock()
	throw (IopcDeadlockException)
{
	int res;
	switch (res = pthread_mutex_lock(&mutex)) {
	case 0:
		return;
	break;
	case EDEADLK:
		throw IopcDeadlockException();
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

void Lock::unlock()
	throw (IopcIncorrectUsageException)
{
	switch (pthread_mutex_unlock(&mutex)) {
	case 0:
		return;
	break;
	case EPERM:
		throw IopcIncorrectUsageException();
	break;
	default:
		IOPC_UNREACHABLE;
	break;
	}
}

}
