/**
 * @file rwxLock.h
 *
 * @author Jan Hadrava
 * @ingroup iopccommon
 */

#ifndef RWXLOCK_H_
#define RWXLOCK_H_

#include <pthread.h>

#include "exceptionBase.h"
#include "mutex.h"
#include "defs.h"

namespace iopc {

/**
 * @brief Recursive read/write/exclusive lock sychronization primitive.
 *
 * Allows several write locks to be held by a thread.
 *
 * @ingroup iopccommon
 */
class DLL_EXTERNAL RWXLock {
public:
	RWXLock()
		throw (IopcNoMemoryException, IopcRsrcLackException);
	~RWXLock()
		throw ();
	/** @brief Waits lock for reading. */
	void rlock()
		throw (IopcRsrcLackException, IopcDeadlockException, IopcNoMemoryException);
	/** @brief Waits for lock writing. */
	void wlock()
		throw (IopcDeadlockException, IopcNoMemoryException);
	/** @brief Waits for exclusive lock (only one is allowed). */
	void xlock()
		throw (IopcDeadlockException, IopcNoMemoryException);
	/** @brief Tries to get lock for reading immediately. */
	bool tryRLock()
		throw (IopcRsrcLackException, IopcDeadlockException);
	/** @brief Tries to get lock for writing immediately. */
	bool tryWLock()
		throw (IopcRsrcLackException, IopcDeadlockException);
	/** @brief Tries to get exclusive lock immediately. */
	bool tryXLock()
		throw (IopcDeadlockException);
	/** @brief Removes one lock. */
	void unlock()
		throw (IopcIncorrectUsageException);
private:
	RWXLock(const RWXLock& );
	RWXLock& operator=(const RWXLock& );

	int wlocks;	/**< Number of locks for writing. */
	bool isXLock;	/**< Indicates whether exclusive lock is locked. */
	pthread_rwlock_t rwlock;
	Mutex mutex;
	pthread_t wlocker;	/**< Owner of locks for writing. */
};

}

#endif /*RWXLOCK_H_*/
