/**
 * @file basicDatabase.cpp
 *
 * @author Michal Kopecky
 * @author Jan Hadrava
 * @author Josef Troch
 * @author Petr Cermak
 *
 * @ingroup iopcdb
 */
#include <string>
#include "basicDatabase.h"
#include "iopccommon/trace.h"
#include "databaseImpl.h"
#include "basicConnection.h"

using namespace std;

namespace iopc {

BasicDatabase::ConnectionSet BasicDatabase::allConnections;
Lock BasicDatabase::allLock;

BasicDatabase::BasicDatabase(DatabaseImpl* implementation, Driver& driver) : driver(driver)
{
	IOPC_FUNC_INVOKED;
	this->implementation = implementation;
}

BasicDatabase::~BasicDatabase()
//Database object destructor
{
	IOPC_FUNC_INVOKED;
	if (connections.size() > 0) {
		LogWriter::logWarning("Removing registered connections.");
		disconnect();
	}
}

Database* BasicDatabase::getBasicDatabase()
{
	return this;
}

void BasicDatabase::addConnection(Connection* con)
{
	allLock.lock();
	try {
		allConnections.insert(con);
		connections.insert(con);
	}
	IOPC_LOG_CATCH(...)
		allLock.unlock();
		IOPC_LOG_RETHROW(...);
	IOPC_LOG_CAUGHT
	allLock.unlock();
}

void BasicDatabase::removeConnection(Connection* con)
{
	IOPC_FUNC_INVOKED;
	allLock.lock();
	try {
		ConnectionSetIterator iter = allConnections.find(con);
		if (iter != allConnections.end())
			allConnections.erase(iter);

		iter = connections.find(con);
		if (iter != connections.end())
			connections.erase(iter);
	}
	IOPC_LOG_CATCH(...)
		allLock.unlock();
		IOPC_LOG_RETHROW(...);
	IOPC_LOG_CAUGHT
	allLock.unlock();
	IOPC_FUNC_ENDED;
}

Connection* BasicDatabase::getConnection(const string& connStr) {
	Connection* conn = new BasicConnection(implementation->getConnection(connStr), *this);
	lock.lock();
	try {
		addConnection(conn);
	}
	IOPC_LOG_CATCH(...)
		lock.unlock();
		IOPC_LOG_RETHROW(...);
	IOPC_LOG_CAUGHT
	lock.unlock();
	return conn;
}

void BasicDatabase::returnConnection(Connection* connection) {
	IOPC_FUNC_INVOKED;
	// TODO: vraceni connection co sem nepatri..
	lock.lock();
	try {
		if (connection->isOpened()) {
			connection->close();
		}
		removeConnection(connection->getBasicConnection());
		implementation->returnConnection(connection->getImplementation());
	}
	IOPC_LOG_CATCH(...)
		lock.unlock();
		IOPC_LOG_RETHROW(...);
	IOPC_LOG_CAUGHT
	lock.unlock();
	delete connection;
	IOPC_FUNC_ENDED;
}


void BasicDatabase::commit()
{
	IOPC_FUNC_INVOKED;
	lock.lock();
	try {
		for (ConnectionSetIterator iter = connections.begin(); iter != connections.end(); ++iter) {
			(*iter)->commit();
		}
	}
	IOPC_LOG_CATCH(...)
		lock.unlock();
		IOPC_LOG_RETHROW(...);
	IOPC_LOG_CAUGHT
	lock.unlock();
}

void BasicDatabase::rollback()
{
	IOPC_FUNC_INVOKED;
	lock.lock();
	try {
		for (ConnectionSetIterator iter = connections.begin(); iter != connections.end(); ++iter) {
			(*iter)->rollback();
		}
	}
	IOPC_LOG_CATCH(...)
		lock.unlock();
		IOPC_LOG_RETHROW(...);
	IOPC_LOG_CAUGHT
	lock.unlock();
}

void BasicDatabase::disconnect()
{
	IOPC_FUNC_INVOKED;
	lock.lock();
	try {
		for (ConnectionSetIterator iter = connections.begin(); iter != connections.end(); ++iter) {
			(*iter)->disconnect();
			allConnections.erase(*iter);
			implementation->returnConnection((*iter)->getImplementation());
			delete (*iter);
		}
		connections.clear();
	}
	IOPC_LOG_CATCH(...)
		lock.unlock();
		IOPC_LOG_RETHROW(...);
	IOPC_LOG_CAUGHT
	lock.unlock();
	IOPC_FUNC_ENDED;
}


}
