/**
 * @file driverManager.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopcdb
 */

#ifndef DRIVERMANAGER_H_
#define DRIVERMANAGER_H_

#include "iopccommon/defsPlatform.h"
#include "database.h"
#include "driver.h"
#include "driverExtension.h"
#include <string>
#include <vector>

namespace iopc {

/**
 * @brief A singleton class managing set of IOPC database drivers.
 *
 * Each driver is registered under a name at application startup.
 * A user can later get a reference to a driver by specifying its name
 * when calling the @a getDriver(const std::string& name) static method.
 *
 * This class is threadsafe for driver reference retrieval.
 *
 * @ingroup iopcdb
 */
class DLL_EXTERNAL DriverManager {
	public:
		static DriverManager& getInstance();

		/**
		 * @brief Registers a new driver to the driver manager
		 *
		 * This method is called automatically by a @a Driver::Driver(const std::string& driverName) constructor which is called
		 * during static driver initilaization.
		 *
		 * @param driver A pointer to the driver to be registered (global instance!)
		 * @param name Name of the driver being registered
		 */
		void registerDriver(Driver* driver, const std::string& name);

		/**
		 * @brief Registers a driver extension to a driver
		 *
		 * This method is called automatically by a @a DriverExtension::DriverExtension(const std::string& driverName, const std::string& name)
		 * constructior which is called during static driver extension initialization.
		 *
		 * @param driverEx A pointer to the driver extension to be registered (stored in global variable)
		 * @param driverName Name of a registered driver which the extension to add to
		 * @param name Name of the extension being registered
		 */
		void registerDriverExtension(DriverExtension* driverEx, const std::string& driverName, const std::string& name);

		/**
		 * @brief Gets a driver registered by a name.
		 *
		 * @param name The driver name
		 * @returns Driver identified by the name.
		 */
		Driver& getDriverInst(const std::string& name);

		/**
		 * @brief Runs driver initialization blocks
		 *
		 * This method should be run after static initialization has completed and the main() method has been run.
		 * The method is run automatically from @a IopcDb::init(). The @a Driver::init(std::map<std::string, DriverExtension*>& extensions)
		 * method is called on each registered driver.
		 *
		 * @see IopcDb::init()
		 * @see Driver::init(std::map<std::string, DriverExtension*>& extensions)
		 */
		void initDrivers();

		/**
		 * @brief Runs driver termination blocks
		 *
		 * This method should be run before exiting the program.
		 * The method is run automatically from @a IopcDb::shutdown(). The @a Driver::shutdown()
		 * method is called on each registered driver.
		 *
		 * @see IopcDb::shutdown()
		 * @see Driver::shutdown()
		 */
		void shutdownDrivers();

		/**
		 * @brief Gets a driver registered by a name.
		 *
		 * @param name The driver name
		 * @returns driver identified by the name.
		 */
		inline static Driver& getDriver(const std::string& name) {
			return getInstance().getDriverInst(name);
		}

		/**
		 * @brief A shortcut for @a Driver::returnDatabase(Database* db)
		 *
		 * @see Driver::returnDatabase(Database* db)
		 */
		inline static void returnDatabase(Database* db) {
			db->getDriver().returnDatabase(db);
		}

	protected:
		DriverManager();
		~DriverManager();

	private:
		std::map<std::string, Driver*> drivers;
		std::map<std::string, std::map<std::string, DriverExtension*> > driverExtensions;

		Mutex mutex;

};

}

#endif /*DRIVERMANAGER_H_*/
