/**
 * @file sqlStatementsFeature.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopcdb
 */

#ifndef SQLSTATEMENTSFEATURE_H_
#define SQLSTATEMENTSFEATURE_H_

#include "dbFeature.h"
#include "iopccommon/metadataHolder.h"
#include <vector>

namespace iopc {

/**
 * @brief Declares methods that are used to generate basic SQL statements
 *
 * These statements are:
 * CREATE TABLE, DROP TABLE, INSERT INTO table, UPDATE table, DELETE FROM table, SELECT FROM table
 * CREATE VIEW, DROP VIEW.
 *
 * @ingroup iopcdb
 */
class SqlStatementsFeature : public DbFeature
{
private:
	static std::string featureName;
public:
	virtual const std::string& getNameInst() {
		return getName();
	}

	static const std::string& getName() {
		return featureName;
	}

	/**
	 * @brief Generates a CREATE TABLE statement
	 *
	 * @param tableMeta Metadata describing the table.
	 * <ul>
	 *   <li>[db.table] - Name of the table</li>
	 *   <li>[db.type] - If O/R database is used, a table [db.table] of the ADT type [db.type] is generated</li>
	 * </ul>
	 * @param columns Metadata describing the columns.
	 * <ul>
	 *   <li>[db.column] - Column name</li>
	 *   <li>[db.type.notNull] - Column is rendered as NOT NULL</li>
	 *   <li>[db.primaryKey] - Column is part of the primary key</li>
	 * </ul>
	 * @param createMeta Additional driver-specific parameters
	 */
	virtual std::string getCreateTableStatement(const MetadataHolder& tableMeta, const std::vector<const MetadataHolder*>& columns, const MetadataHolder& createMeta = MetadataHolder::emptyMeta) =0;

	/**
	 * @brief Generates a DROP TABLE statement
	 *
	 * @param tableMeta Metadata describing the table to drop
	 * <ul>
	 * 	<li>[db.table] - Name of the table</li>
	 * </ul>
	 * @param dropmeta Additional driver-specific parameters
	 */
	virtual std::string getDropTableStatement(const MetadataHolder& tableMeta, const MetadataHolder& dropmeta = MetadataHolder::emptyMeta) =0;

	/**
	 * @brief Generates an INSERT statement
	 *
	 * @param tableMeta Metadata describing the table to drop
	 * <ul>
	 * 	<li>[db.table] - Name of the table</li>
	 * </ul>
	 * @param inColumns Metadata describing the table columns
	 * <ul>
	 *  <li>[db.column] - Column name</li>
	 * </ul>
	 * @param insertMeta Additional driver-specific parameters
	 */
	virtual std::string getInsertStatement(const MetadataHolder& tableMeta, const std::vector<const MetadataHolder*>& inColumns, const MetadataHolder& insertMeta = MetadataHolder::emptyMeta) =0;

	/**
	 * @brief Generates an UPDATE statement
	 *
	 * @param tableMeta Metadata describing the table to update
	 * <ul>
	 * 	<li>[db.table] - Name of the table</li>
	 * </ul>
	 * @param inColumns Metadata describing the columns which will be updated
	 * <ul>
	 *  <li>[db.column] - Column name</li>
	 * </ul>
	 * @param updateMeta Additional parameters
	 * <ul>
	 * 	<li>[db.update.where] - WHERE part of the generated UPDATE statement (optional)
	 * </ul>
	 */
	virtual std::string getUpdateStatement(const MetadataHolder& tableMeta, const std::vector<const MetadataHolder*>& inColumns, const MetadataHolder& updateMeta = MetadataHolder::emptyMeta) =0;

	/**
	 * @brief Generates a DELETE statement
	 *
	 * @param tableMeta Metadata describing the table from which to delete rows
	 * <ul>
	 * 	<li>[db.table] - Name of the table</li>
	 * </ul>
	 * @param deleteMeta Additional parameters
	 * <ul>
	 * 	<li>[db.delete.where] - WHERE part of the generated DELETE statement
	 * </ul>
	 */
	virtual std::string getDeleteStatement(const MetadataHolder& tableMeta, const MetadataHolder& deleteMeta = MetadataHolder::emptyMeta) =0;

	/**
	 * @brief Generates a basic SELECT statement
	 *
	 * @param tables Metadata describing the tables from which to select data
	 * <ul>
	 * 	<li>[db.table] - Name of the table</li>
	 * </ul>
	 * @param outColumns Metadata describing the columns which will be selected
	 * <ul>
	 *  <li>[db.column] - Column name</li>
	 *  <li>[db.column.alias] - Alias of the column in the SELECT statement (optional)</li>
	 * </ul>
	 * @param selectMeta Additional parameters
	 * <ul>
	 * 	<li>[db.update.where] - WHERE part of the generated SELECT statement (optional)</li>
	 *  <li>[db.update.groupBy] - GROUP BY part of the generated SELECT statement (optional) </li>
	 *  <li>[db.update.having] - HAVING part of the generated SELECT statement (optional) </li>
	 *  <li>[db.update.oraderBy] - ORDER BY part of the generated SELECT statement (optional) </li>
	 *  <li>[db.select.lock] - boolean, determines whether the generated select should lock the rows returned (FOR UPDATE) (optional)</li>
	 *  <li>[db.select.nowait] - boolean, determines whether the generated select should wait for the required locks to be freed or not (NOWAIT) (optional) </li>
	 * </ul>
	 */
	virtual std::string getSelectStatement(const std::vector<const MetadataHolder*>& tables, const std::vector<const MetadataHolder*>& outColumns, const MetadataHolder& selectMeta = MetadataHolder::emptyMeta) =0;

	/**
	 * @brief Generates a CREATE VIEW statement
	 * @param meta Metadata describing the view
	 * <ul>
	 *  <li>[db.view] - View name</li>
	 *  <li>[db.view.select] - View code</li>
	 * </ul>
	 * @param createMeta Additional driver-specific parameters
	 */
	virtual std::string getCreateViewStatement(const MetadataHolder& meta, const MetadataHolder& createMeta = MetadataHolder::emptyMeta) =0;

	/**
	 * @brief Generates a DROP VIEW statement
	 * @param meta MEtadata describing the view
	 * <ul>
	 *  <li>[db.view] - View name</li>
	 * </ul>
	 * @param dropMeta Additional driver-specific parameters
	 */
	virtual std::string getDropViewStatement(const MetadataHolder& meta, const MetadataHolder& dropMeta = MetadataHolder::emptyMeta) =0;

	/**
	 * @brief Generates a CREATE INDEX statement
	 * @param tableMeta Metadata describing the table on which to create the index
	 * <ul>
	 *  <li>[db.table] - Name of the table</li>
	 * </ul>
	 * @param columns Metadata describing the indexes
	 * <ul>
	 * 	<li>[db.index] - Name of the index</li>
	 * 	<li>[db.index.unique] - Name of the unique index (at least one of these two metadata must be specified)</li>
	 * 	<li>[db.column] - Name of the column on which to create the index</li>
	 * </ul>
	 */
	virtual std::vector<std::string> getIndexes(const MetadataHolder& tableMeta, const std::vector<const MetadataHolder*>& columns) =0;
};


}

#endif /*SQLSTATEMENTSFEATURE_H_*/
