/**
 * @file arcCacheItem.cpp
 *
 * @author Jan Hadrava
 * @date 2004
 * @ingroup iopclibcache
 */

#include "arcCacheItem.h"
#include "arcCacheBase.h"

#include "cache.h"
#include "iopcdb/connection.h"

namespace iopc {

ArcCacheItem::ArcCacheItem()
: CacheLock(), lock(),
	cache(NULL), strat(), persId(), objectPtr(NULL), state(ArcCacheItem::ST_BORN),
	next(this), prev(this), isLocked(false), timeStamp(), frequent(false)
{
}

ArcCacheItem::ArcCacheItem(ArcCacheBase &cache, const Strategy &strat, const PersIdentification persId, DatabaseObject *obj, State state)
: CacheLock(), lock(),
	cache(&cache), strat(strat), persId(persId), objectPtr(obj), state(state),
	next(NULL), prev(NULL), isLocked(false), timeStamp(), frequent(false)
{
}

ArcCacheItem::~ArcCacheItem()
{
	IOPC_ASSERT((prev == NULL) == (next == NULL));
	if (prev != NULL) {
		IOPC_ASSERT(next != NULL);
		prev->next = next;
		next->prev = prev;
	}
	try {
		lock.xlock();	/* wait for all locks to be unlocked */
		lock.unlock();
		if (objectPtr != NULL) {
			delete objectPtr;
		}
	} catch (...) {
		IOPC_UNREACHABLE;
	}
}

DatabaseObject& ArcCacheItem::getObject()
{
	if (objectPtr == NULL) {
		IOPC_LOG_THROW(IopcIncorrectUsageException());
	}
	return *objectPtr;
}

CacheLock& ArcCacheItem::addLock(bool rdonly)
{
	return cache->addLock(*this, rdonly);
}

void ArcCacheItem::rmLock()
{
	/* this can require some change in cache queues */
	cache->rmLock(*this);
}

void ArcCacheItem::replaceObject(DatabaseObject* obj)
{
	if (objectPtr != NULL) {
		delete objectPtr;
	}
	objectPtr = obj;
}

void ArcCacheItem::remove()
{
	IOPC_ASSERT((prev == NULL) == (next == NULL));
	if (prev != NULL) {
		prev->next = next;
		next->prev = prev;
		prev = next = NULL;
	}
}

void ArcCacheItem::addNext(ArcCacheItem& item)
{
	next->prev = &item;
	item.next = next;
	item.prev = this;
	next = &item;
}

void ArcCacheItem::addPrev(ArcCacheItem &item)
{
	prev->next = &item;
	item.prev = prev;
	item.next = this;
	prev = &item;
}

void ArcCacheItem::release()
{
	if (objectPtr != NULL) {
		delete objectPtr;
		objectPtr = NULL;
	}
}

void ArcCacheItem::setFrequent(bool frequent)
{
	this->frequent = frequent;
}

bool ArcCacheItem::isFrequent()
{
	return frequent;
}

}/*namespace iopc*/
