/**
 * @file arcCacheST.h
 *
 * @author Jan Hadrava
 * @date 2004
 * @ingroup iopclibcache
 */

#ifndef ARCCACHEST_H_
#define ARCCACHEST_H_



#include <map>

#ifndef WIN32
#include <ext/hash_map>
#endif /*WIN32*/

#include "iopccommon/defs.h"
#include "arcCacheBase.h"
#include "arcCacheItem.h"
#include "types.h"
#include "iopccommon/condVar.h"


namespace iopc {

template<typename MapType> class ArcCacheST;

typedef ArcCacheST<std::map<PersIdentification, ArcCacheItem *> > MapArcCacheST;

#ifndef WIN32
typedef ArcCacheST<__gnu_cxx::hash_map<PersIdentification, ArcCacheItem *> > HashArcCacheST;
#endif /*WIN32*/

/** Singlethreaded cache using ARC replacement policy. Objects are represented by ArcCacheItem.
* Items are stored an container conforming to STL "pair associative container" concept. Its type
* is specified by template parameter.
*
* @ingroup iopclibcache
*/
template<typename MapType> class ArcCacheST: public ArcCacheBase
{
private:
#ifndef WIN32
	typedef class MapType::iterator Iter;	// MSVC6 doesn't like this although it's perfectly conforming to C++ standard
#else
	typedef MapType::iterator Iter;		// gcc doesn't like this
#endif /*WIN32*/
public:
	inline ArcCacheST(long minValue = MINLONG, long maxCost = MAXLONG, const Strategy& strat = Strategy::defaultStrategy);
	inline ArcCacheST(long minValue, long maxCost, StrategySelector* stratSel);
	inline virtual ~ArcCacheST();
private:/*hiding*/
	ArcCacheST(const ArcCacheST<MapType>& );
	ArcCacheST<MapType> &operator=(const ArcCacheST<MapType>& );
protected:
	virtual void clear(const StrategySelector& strSel);

	virtual CacheLock& makePersistent(const StrategySelector& strSel, const PersIdentification& persId, DatabaseObject& objectPtr);
	virtual CacheLock& getCacheLock(const StrategySelector& strSel, const PersIdentification& persId, bool rdonly = false);

	virtual void dbDelete(const StrategySelector& strSel, const PersIdentification& persId);
	virtual void cacheDelete(const StrategySelector& strSel, const PersIdentification& persId);

	virtual void updateDirty(const StrategySelector& strSel);
	virtual void removeAll(const StrategySelector& strSel);
protected:
	ArcCacheItem& addLock(ArcCacheItem& item, bool rdonly);
	void rmLock(ArcCacheItem& item);
public:
	virtual inline long getActiveCost();
	virtual inline long getLockedCost();
	virtual long getWorstValue(const TimeStamp& current);
	virtual long removeWorst();
	virtual bool maintainerStep(long& costChange);
	virtual long getSleepTime(const TimeStamp& current, long minValue);
	inline virtual Cache* clone() {
		return new ArcCacheST<MapType>(*this);
	}
protected:
	/** Remove one object scheduled for deletion.
	* \return false there are no objects scheduled for deletion, true otherwise.
	*/
	virtual bool removeDoomed();
	/** Removes one zombie. Zombies are created when an item can't removed
	* from container because of the danger of deadlock.
	* \return false when there are no zombies present, true otherwise.
	*/
	virtual bool removeZombie();
	/** Updates one modified unlocked object.
	* \param costChange Output parameter, the change of total cost caused by method call.
	* \return false when there are no modified unlocked objects present, true otherwise.
	*/
	virtual bool asyncUpdate(long& costChange);
protected:
	MapType cont;		/**< Container holding items of all types - locked/unlocked, representing scheduled actions, modified, history, ... */
	long rActiveCnt, rLockedCnt, fActiveCnt, fLockedCnt, rHistCnt, fHistCnt;
	ArcCacheItem rActiveList;
	ArcCacheItem rHistList;
	ArcCacheItem fActiveList;
	ArcCacheItem fHistList;
	ArcCacheItem doomedList;	/* double linked list of items scheduled for deletion */
	ArcCacheItem zombieList;	/* double linked list of item remnants */
	long activeCost, lockedCost;	/* doomedCost == 0 && zombieCost == 0 */
	double target;
};

template<typename MapType> ArcCacheST<MapType>::ArcCacheST(long minValue, long maxCost, StrategySelector* stratSel)
: ArcCacheBase(minValue, maxCost, stratSel),
	rActiveCnt(0), rLockedCnt(0), fActiveCnt(0), fLockedCnt(0), rHistCnt(0), fHistCnt(0),
	rActiveList(), rHistList(), fActiveList(), fHistList(),
	doomedList(), zombieList(),
	activeCost(0), lockedCost(0), target(0.5)
{
}

template<typename MapType> ArcCacheST<MapType>::ArcCacheST(long minValue, long maxCost, const Strategy& strat)
: ArcCacheBase(minValue, maxCost, strat),
	rActiveCnt(0), rLockedCnt(0), fActiveCnt(0), fLockedCnt(0), rHistCnt(0), fHistCnt(0),
	rActiveList(), rHistList(), fActiveList(), fHistList(),
	doomedList(), zombieList(),
	activeCost(0), lockedCost(0), target(0.5)
{
}

template<typename MapType> ArcCacheST<MapType>::ArcCacheST(const ArcCacheST<MapType>& src)
: ArcCacheBase(src.minValue, src.maxCost, src.stratSel, src.defaultStrategy),
	rActiveCnt(0), rLockedCnt(0), fActiveCnt(0), fLockedCnt(0), rHistCnt(0), fHistCnt(0),
	rActiveList(), rHistList(), fActiveList(), fHistList(),
	doomedList(), zombieList(),
	activeCost(0), lockedCost(0), target(0.5)
{
}

template<typename MapType> ArcCacheST<MapType>::~ArcCacheST()
{
	ExtendedCache::clear();
}

template<typename MapType> void ArcCacheST<MapType>::clear(const StrategySelector& strSel)
{

	long costChange = 0;
	for (Iter iter = cont.begin();  iter != cont.end(); iter++) {
		ArcCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */

		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
				/* avoiding deadlock */
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		long cost;
		switch (item->state) {
		case ArcCacheItem::ST_BORN:	/* new item */
			if (item->getObject().isDirty()) {
				try {
					persId.insert(*conn, item->getObject());
				}
				IOPC_LOG_CATCH(IopcException& exc)
						/* avoiding deadlock */
					IOPC_LOG_THROW(exc);
				IOPC_LOG_CATCH_NEXT(...)
					IOPC_UNREACHABLE;
				IOPC_LOG_CAUGHT
			}

			item->remove();
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;

		break;
		case ArcCacheItem::ST_ADULT:	/* item with DB copy */
			if (item->getObject().isDirty()) {
				try {
					persId.update(*conn, item->getObject());
				}
				IOPC_LOG_CATCH(IopcException& exc)
						/* avoiding deadlock */
					IOPC_LOG_THROW(exc);
				IOPC_LOG_CATCH_NEXT(...)
					IOPC_UNREACHABLE;
				IOPC_LOG_CAUGHT
			}

			item->remove();
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;

		break;
		case ArcCacheItem::ST_DOOMED:	/* item scheduled for deletion */
			try {
				persId.remove(*conn);
			}
			IOPC_LOG_CATCH(IopcException& exc)
					/* avoiding deadlock */
				IOPC_LOG_THROW(exc);
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT

			item->remove();

		break;
		case ArcCacheItem::ST_ZOMBIE:	/* zombie item (busy items can be converted to zombie) */

			item->remove();

		break;
		case ArcCacheItem::ST_HIST:

			item->remove();
			item->isFrequent() ? fHistCnt-- : rHistCnt--;

		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}

		delete item;
	}
	cont.clear();

//	IOPC_ASSERT(fActiveCnt == 0 && rActiveCnt == 0 && fLockedCnt == 0 && rLockedCnt == 0);
	IOPC_ASSERT(activeCost == 0 && lockedCost == 0);
	IOPC_ASSERT(rActiveList.prev == &rActiveList && rActiveList.next == &rActiveList);
	IOPC_ASSERT(fActiveList.prev == &fActiveList && fActiveList.next == &fActiveList);
	IOPC_ASSERT(rHistList.prev == &rHistList && rHistList.next == &rHistList);
	IOPC_ASSERT(fHistList.prev == &fHistList && fHistList.next == &fHistList);
	IOPC_ASSERT(doomedList.prev == &doomedList && doomedList.next == &doomedList);
	IOPC_ASSERT(zombieList.prev == &zombieList && zombieList.next == &zombieList);
	const Strategy& strat = strSel.select();

	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> CacheLock &ArcCacheST<MapType>::makePersistent(const StrategySelector& strSel, const PersIdentification& persId, class DatabaseObject &objectPtr)
{
	long costChange;

	const Strategy& strat = strSel.select(persId);
	ArcCacheItem* item;
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		item = iter->second;
		/* wait for other thread to finish its work */

		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		switch (item->state) {
		case ArcCacheItem::ST_BORN:
		case ArcCacheItem::ST_ADULT:

			IOPC_LOG_THROW(IopcDuplicateKeyException());
		default:
			IOPC_UNREACHABLE;
		break;
		}
		ArcCacheItem::State oldstate = item->state;
		item->state = ArcCacheItem::ST_BUSY;	/* we will keep the item in the cache */

		switch (oldstate) {
		case ArcCacheItem::ST_BORN:
		case ArcCacheItem::ST_ADULT:
			IOPC_UNREACHABLE;
		break;
		case ArcCacheItem::ST_HIST:
			/* this is strange - item with the same ID as the one being inserted was recently evicted from cache */

			item->remove();	/* remove from histList */
			item->isFrequent() ? fHistCnt-- : rHistCnt--;

		break;
		case ArcCacheItem::ST_DOOMED:
			try {
				persId.remove(*conn);	/* doing maintainer's work */
			}
			IOPC_LOG_CATCH(IopcException& exc)

				item->state = ArcCacheItem::ST_DOOMED;


				IOPC_LOG_THROW(exc);
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT
		/*nobreak;*/
		case ArcCacheItem::ST_ZOMBIE:

			item->remove();

		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		/* now we have in cache busy item which is prepared to become born/adult */
	} else {
		item = new ArcCacheItem(*this, strat, persId, NULL, ArcCacheItem::ST_BUSY);
		cont[persId] = item;

	}
	IOPC_ASSERT(item->state == ArcCacheItem::ST_BUSY);
	UpdatingStrategy uStrat = strat.getUpdatingStrategy();
	if (uStrat == USImmed && strat.getLoadLocking()) {
		uStrat = USDeferred;	/* it's not possible to immediately insert object which should be locked in database */
	}
	if (uStrat == USImmed) {
		objectPtr.setDirty(false);
	} else {
		objectPtr.setDirty(true);
	}
	if (uStrat == USImmed) {
		try {
			persId.insert(*conn, objectPtr);
		}
		IOPC_LOG_CATCH(IopcException& exc)

			zombieList.addPrev(*item);

			costChange = 0;

			item->state = ArcCacheItem::ST_ZOMBIE;


			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT

		objectPtr.setDirty(false);
	}
	item->isLocked = true;
	item->setFrequent(false);
	item->addLock(false);
	//objectPtr.SetPersistent(*conn);
	item->objectPtr = &objectPtr;

	item->state = uStrat == USImmed ? ArcCacheItem::ST_ADULT : ArcCacheItem::ST_BORN;




	rLockedCnt++;
	lockedCost += costChange = strat.cost(item->getObject());


	strat.wakeUpMaintainer(costChange);
	imposeLimits();
	return *item;
}

template<typename MapType> CacheLock &ArcCacheST<MapType>::getCacheLock(const StrategySelector& strSel, const PersIdentification& persId, bool rdonly)
{

	const Strategy& strat = strSel.select(persId);
	ArcCacheItem* item;
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		item = iter->second;
		/* wait for other thread to finish its work */

		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		switch (item->state) {
		case ArcCacheItem::ST_BORN:
		case ArcCacheItem::ST_ADULT: {
		/* item is present and active */
			item->addLock(rdonly);

			if (item->isLocked) {
				/* item was already locked,
				** so there's no need to remove it to cleanList
				** and it will be unsafe to reload it */


				strat.wakeUpMaintainer(0);
				imposeLimits();
				return *item;
			}
			/* first lock added, item must be added to lockedList and can be reloaded */
			item->remove(); /* removes item from activeList or updateList */
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			activeCost -= strat.cost(item->getObject());
			fLockedCnt++;
			lockedCost += strat.cost(item->getObject());
			item->setFrequent(true);
			item->isLocked = true;

			ReadingStrategy rStrat = strat.getReadingStrategy();
			if (item->state == ArcCacheItem::ST_BORN || rStrat == RSCache || item->getObject().isDirty()) {
				/* there's no need to reload object */

				strat.wakeUpMaintainer(0);
				imposeLimits();
				return *item;
			}
			/* object will be reloaded */
			IOPC_ASSERT(item->state == ArcCacheItem::ST_ADULT);
			item->state = ArcCacheItem::ST_BUSY;
			 /* avoid blocking whole cache during reloading */
			try {
				switch (rStrat) {
				case RSTimestamp:
					if (persId.versionMatches(*conn, dynamic_cast<OidObject&>(item->getObject()))) {
						break;
					}
				/*nobreak*/
				case RSDatabase: {
					DatabaseObject* objectPtr = persId.load(*conn, strat.getLoadWait(), strat.getLoadLocking());
					IOPC_ASSERT(objectPtr != NULL);
					item->replaceObject(objectPtr);
				break; }
				case RSCache:
				break;
				default:
					IOPC_UNREACHABLE;
				break;
				}
			}
			IOPC_LOG_CATCH(IopcException&)	/* reload failed, ignore it */
			IOPC_LOG_CATCH_NEXT(...)
				IOPC_UNREACHABLE;
			IOPC_LOG_CAUGHT

			item->state = ArcCacheItem::ST_ADULT;


			strat.wakeUpMaintainer(0);
			imposeLimits();
			return *item;
		break; }
		case ArcCacheItem::ST_DOOMED:
		/* item is on the verge of extinction */

			IOPC_LOG_THROW(IopcNotFoundException("Item not present"));
		break;
		case ArcCacheItem::ST_ZOMBIE:

			item->remove();

			item->setFrequent(false);
		break;
		case ArcCacheItem::ST_HIST:

			/* adaptation */
			if (item->isFrequent()) {
				IOPC_ASSERT(fHistCnt > 0);
				long fCnt = fActiveCnt + fLockedCnt;
				if (fCnt == 0) {
					fCnt = 1;
				}
				target -= ((0.0 + rHistCnt)/fHistCnt)/(fCnt);
				if (target < 0.0) {
					target = 0.0;
				}
			} else {
				IOPC_ASSERT(rHistCnt > 0);
				long fCnt = fActiveCnt + fLockedCnt;
				if (fCnt == 0) {
					fCnt = 1;
				}
				target += ((0.0 + fHistCnt)/rHistCnt)/(fCnt);
				if (target > 1.0) {
					target = 1.0;
				}
			}
			item->remove();
			item->isFrequent() ? fHistCnt-- : rHistCnt--;

			item->setFrequent(true);
		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		item->state = ArcCacheItem::ST_BUSY;
	} else {
		item = new ArcCacheItem(*this, strat, persId, NULL, ArcCacheItem::ST_BUSY);
		item->setFrequent(false);
		cont[persId] = item;
	}

	DatabaseObject* objectPtr;
	try {
		objectPtr = persId.load(*conn, strat.getLoadWait(), strat.getLoadLocking());
	}
	IOPC_LOG_CATCH(IopcException& exc)

		zombieList.addPrev(*item);


		item->state = ArcCacheItem::ST_ZOMBIE;


		IOPC_LOG_THROW(exc);
	IOPC_LOG_CATCH_NEXT(...)
		IOPC_UNREACHABLE;
	IOPC_LOG_CAUGHT
	IOPC_ASSERT(objectPtr != NULL);
	//objectPtr->SetPersistent(*conn);
	item->objectPtr = objectPtr;
	item->addLock(rdonly);
	item->isLocked = true;

	long costChange;
	item->isFrequent() ? fLockedCnt++ : rLockedCnt++;
	lockedCost += costChange = strat.cost(item->getObject());


	item->state = ArcCacheItem::ST_ADULT;



	strat.wakeUpMaintainer(costChange);
	imposeLimits();
	return *item;
}

template<typename MapType> void ArcCacheST<MapType>::dbDelete(const StrategySelector& strSel, const PersIdentification& persId)
{
	long costChange = 0;

	const Strategy& strat = strSel.select(persId);
	ArcCacheItem* item;
	Iter iter = cont.find(persId);
	if (iter != cont.end()) {
		item = iter->second;
		/* wait for other thread to finish its work */

		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)

			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		switch (item->state) {
		case ArcCacheItem::ST_BORN:	/* remove it immediately */

			item->remove();
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			activeCost += costChange = -strat.cost(item->getObject());

			cont.erase(iter);

			delete item;
			strat.wakeUpMaintainer(costChange);
			imposeLimits();
			return; /* ignoring the possibility an item already present in the database
				** can have the same key as the removed one */
		break;
		case ArcCacheItem::ST_DOOMED:
		/* item is on the verge of extinction */

			IOPC_LOG_THROW(IopcNotFoundException("Item not present"));
		break;
		case ArcCacheItem::ST_ADULT:

			item->remove();
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			activeCost += costChange = -strat.cost(item->getObject());

			item->release();
			item->state = ArcCacheItem::ST_BUSY;

		break;
		case ArcCacheItem::ST_ZOMBIE:

			item->remove();
			item->state = ArcCacheItem::ST_BUSY;

			item->release();	/* release data */

			costChange = 0;
		break;
		case ArcCacheItem::ST_HIST:

			item->remove();
			item->isFrequent() ? fHistCnt-- : rHistCnt--;
			item->state = ArcCacheItem::ST_BUSY;

			item->release();	/* release data */

			costChange = 0;
		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		//IOPC_UNREACHABLE;
	} else {
		item = new ArcCacheItem(*this, strat, persId, NULL, ArcCacheItem::ST_BUSY);
		cont[persId] = item;

		costChange = 0;
	}
	IOPC_ASSERT(item->state == ArcCacheItem::ST_BUSY);
	if (strat.getUpdatingStrategy() == USImmed) {

		zombieList.addPrev(*item);

		try {
			persId.remove(*conn);
		}
		IOPC_LOG_CATCH(IopcException& exc)

			item->state = ArcCacheItem::ST_ZOMBIE;


			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT

		item->state = ArcCacheItem::ST_ZOMBIE;


	} else { /* USAsync || USImmedDeferred */

		doomedList.addPrev(*item);	/* schedule for deletion */


		item->state = ArcCacheItem::ST_DOOMED;


	}
	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> void ArcCacheST<MapType>::cacheDelete(const StrategySelector& strSel, const PersIdentification& persId)
{
	/* ignoring user's advices */
}

template<typename MapType> void ArcCacheST<MapType>::updateDirty(const StrategySelector& strSel)
{
	Iter next;

	long costChange = 0;
	for (Iter iter = cont.begin(); iter != cont.end(); iter = next) {
		next = iter;
		next++;
		ArcCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */

		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
				/* avoiding deadlock */
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();
		switch (item->state) {
		case ArcCacheItem::ST_BORN:	/* new item */
			IOPC_ASSERT(item->getObject().isDirty());
			persId.insert(*conn, item->getObject());
			item->getObject().setDirty(false);
			item->state = ArcCacheItem::ST_ADULT;
		break;
		case ArcCacheItem::ST_ADULT:	/* item with DB copy */
			if (item->getObject().isDirty()) {
				persId.update(*conn, item->getObject());
				item->getObject().setDirty(false);
			}
		break;
		case ArcCacheItem::ST_DOOMED:	/* item scheduled for deletion */
			persId.remove(*conn);

			item->remove();

			cont.erase(iter);
			delete item;
		break;
		case ArcCacheItem::ST_ZOMBIE:	/* zombie item (busy items can be converted to zombie) */

			item->remove();

			cont.erase(iter);
			delete item;
		break;
		case ArcCacheItem::ST_HIST:
			/* keep history */
		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}
		if ((item->state == ArcCacheItem::ST_BORN || item->state == ArcCacheItem::ST_ADULT)
                        && strat.getLoadLocking())
                {       /* item with locking must be removed */

			item->remove();
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			long cost = strat.cost(item->getObject());
			activeCost -= cost;
			costChange -= cost;

			delete item;
			cont.erase(iter);
		}
	}
	IOPC_ASSERT(doomedList.next == &doomedList && doomedList.prev == &doomedList);
	IOPC_ASSERT(zombieList.next == &zombieList && zombieList.prev == &zombieList);


	strSel.select().wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> void ArcCacheST<MapType>::removeAll(const StrategySelector& strSel)
{
	long costChange = 0;

	for (Iter iter = cont.begin(); iter != cont.end(); iter++) {
		ArcCacheItem* item = iter->second;
		PersIdentification& persId = item->persId;
		const Strategy& strat = strSel.select(persId);
		/* wait for other thread to finish its work */

		IOPC_ASSERT(item->state != ArcCacheItem::ST_BUSY);
		/* now we can be sure item's state won't change until contLock is unlocked */
		try {
			item->lock.xlock();	/* wait for the item to be unlocked */
		}
		IOPC_LOG_CATCH(IopcException& exc)
				/* avoiding deadlock */
			IOPC_LOG_THROW(exc);
		IOPC_LOG_CATCH_NEXT(...)
			IOPC_UNREACHABLE;
		IOPC_LOG_CAUGHT
		item->lock.unlock();

		item->remove();
		switch (item->state) {
		case ArcCacheItem::ST_BORN:
		case ArcCacheItem::ST_ADULT:
			item->isFrequent() ? fActiveCnt-- : rActiveCnt--;
			activeCost += costChange = -strat.cost(item->getObject());
		break;
		case ArcCacheItem::ST_DOOMED:
		case ArcCacheItem::ST_ZOMBIE:
		case ArcCacheItem::ST_HIST:
		break;
		case ArcCacheItem::ST_BUSY:
		default:
			IOPC_UNREACHABLE;
		break;
		}

		delete item;
	}
	cont.clear();
	IOPC_ASSERT(rActiveCnt == 0 && rLockedCnt == 0);
	IOPC_ASSERT(fActiveCnt == 0 && fLockedCnt == 0);
	IOPC_ASSERT(activeCost == 0 && lockedCost == 0);
	IOPC_ASSERT(rActiveList.prev == &rActiveList && rActiveList.next == & rActiveList);
	IOPC_ASSERT(fActiveList.prev == &fActiveList && fActiveList.next == & fActiveList);
	IOPC_ASSERT(doomedList.prev == &doomedList && doomedList.next == & doomedList);
	IOPC_ASSERT(zombieList.prev == &zombieList && zombieList.next == & zombieList);


	strSel.select().wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> ArcCacheItem &ArcCacheST<MapType>::addLock(ArcCacheItem& item, bool rdonly)
{
	if (rdonly) {
		item.lock.rlock();
	} else {
		item.lock.wlock();
	}
	return item;
}

template<typename MapType> void ArcCacheST<MapType>::rmLock(ArcCacheItem& item)
{
	/* it's important not to use contLock here so that we avoid deadlock */
	/* it's important not to use contLock here so that we avoid deadlock */
	try {
		item.lock.unlock();
	}
	IOPC_LOG_CATCH(IopcException& exc)
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CATCH_NEXT(...)
		IOPC_UNREACHABLE;
	IOPC_LOG_CAUGHT

	if (!item.isLocked) {

		return;
	}
	if (!item.lock.tryXLock()) {

		return;
	}
	item.remove();	/* remove from list of locked items */
	long cost = item.strat.cost(item.getObject());
	long costChange = 0;
	if (item.isFrequent()) {
		fLockedCnt--;
		fActiveCnt++;
	} else {
		rLockedCnt--;
		rActiveCnt++;
	}
	lockedCost -= cost;
	activeCost += cost;
	item.isLocked = false;
	UpdatingStrategy uStrat = item.strat.getUpdatingStrategy();
	if (uStrat == USImmed && item.state == ArcCacheItem::ST_BORN) {
		IOPC_ASSERT(item.strat.getLoadLocking());
		/* this combination will require inserting item into db and removing local copy but we can't do it
		because it requires state change which can't be made without contLock */
		uStrat = USAsync;
	}
	switch (uStrat) {
	case USDeferred:
		item.isFrequent() ? fActiveList.addNext(item) : rActiveList.addNext(item);
	break;
	case USImmed:
		if (item.state == ArcCacheItem::ST_BORN) {
			IOPC_UNREACHABLE;
		} else {
			IOPC_ASSERT(item.state == ArcCacheItem::ST_ADULT);
			if (item.getObject().isDirty()) {
					/* this is safe because we have xlock for the item */
				item.persId.update(*conn, item.getObject());
				item.getObject().setDirty(false);

			}
			item.isFrequent() ? fActiveList.addNext(item) : rActiveList.addNext(item);
		}
	break;
	case USAsync:
		item.isFrequent() ? fActiveList.addNext(item) : rActiveList.addNext(item);

	break;
	}
	item.timeStamp = TimeStamp::now();
	Strategy strat = item.strat;
	item.lock.unlock();

	strat.wakeUpMaintainer(costChange);
	imposeLimits();
}

template<typename MapType> inline long ArcCacheST<MapType>::getActiveCost()
{
	return activeCost;
}

template<typename MapType> inline long ArcCacheST<MapType>::getLockedCost()
{
	return lockedCost;
}

template<typename MapType> long ArcCacheST<MapType>::getWorstValue(const TimeStamp& time)
{
	long retval = MAXLONG;
	if (rActiveList.prev != &rActiveList) {
		retval = rActiveList.prev->timeStamp - time;	/* difference in milliseconds */
	}
	if (fActiveList.prev != &fActiveList) {
		long value = fActiveList.prev->timeStamp - time;
		if (value < retval) {
			retval = value;
		}
	}
	return retval;

}

template<typename MapType> long ArcCacheST<MapType>::removeWorst()
{


	//ArcCacheItem* list = NULL;
	long value;
	TimeStamp now = TimeStamp::now();
	ArcCacheItem* item = NULL;
	ArcCacheItem* activeList;
	bool freq;
	long cost = 0;
	if (fActiveCnt + fLockedCnt == 0 || (rActiveCnt + rLockedCnt)/(fActiveCnt + fLockedCnt) > target) {
		freq = false;
		activeList = &rActiveList;
	} else {
		freq = true;
		activeList = &fActiveList;
	}
	if (activeList->prev != activeList) {
		value = activeList->prev->timeStamp - now;
		item = activeList->prev;
	}
	if (item != NULL) {
		if (item->state == ArcCacheItem::ST_BUSY) {
			IOPC_LOG_THROW(IopcIncorrectUsageException("Incorrect behavior of maintainer"));
		}
		bool isAdult;
		if (item->state == ArcCacheItem::ST_ADULT) {
			isAdult = true;
		} else {
			IOPC_ASSERT(item->state == ArcCacheItem::ST_BORN);
			isAdult = false;
		}
		item->state = ArcCacheItem::ST_BUSY;
		item->remove();
		//cost;
		activeCost += cost = -item->strat.cost(item->getObject());
		if (freq) {
			fActiveCnt--;
			fHistCnt++;
			fHistList.addNext(*item);
		} else {
			rActiveCnt--;
			rHistCnt++;
			rHistList.addNext(*item);
		}


		if (item->getObject().isDirty()) {
			if (isAdult) {
				item->persId.update(*conn, item->getObject());
			} else {
				item->persId.insert(*conn, item->getObject());
			}
			item->getObject().setDirty(false);
		}
		item->release();

		item->state = ArcCacheItem::ST_HIST;


	}




	while (fHistCnt > rActiveCnt + rLockedCnt && fHistList.prev != &fHistList) {
		item = fHistList.prev;
		item->remove();
		item->release();
		Iter iter = cont.find(item->persId);
		IOPC_ASSERT(iter != cont.end());
		cont.erase(iter);
		delete item;
		fHistCnt--;
	}

	while (rHistCnt > fActiveCnt + fLockedCnt && rHistList.prev != &rHistList) {
		item = rHistList.prev;
		item->remove();
		item->release();
		Iter iter = cont.find(item->persId);
		IOPC_ASSERT(iter != cont.end());
		cont.erase(iter);
		delete item;
		rHistCnt--;
	}




	return cost;
}

template<typename MapType> bool ArcCacheST<MapType>::removeDoomed()
{


	if (doomedList.prev == &doomedList) {


		return false;
	}
	ArcCacheItem& item = *doomedList.prev;

	IOPC_ASSERT(item.state == ArcCacheItem::ST_DOOMED);
	item.state = ArcCacheItem::ST_BUSY;
	item.remove();
	zombieList.addPrev(item);


	item.persId.remove(*conn);

	item.state = ArcCacheItem::ST_ZOMBIE;


	return true;
}

template<typename MapType> bool ArcCacheST<MapType>::removeZombie()
{


	if (zombieList.next == &zombieList) {


		return false;
	}
	ArcCacheItem* item = zombieList.next;

	IOPC_ASSERT(item->state == ArcCacheItem::ST_ZOMBIE);
	Iter iter = cont.find(item->persId);
	IOPC_ASSERT(iter != cont.end());
	cont.erase(iter);
	item->remove();


	delete item;
	return true;
}

template<typename MapType> bool ArcCacheST<MapType>::asyncUpdate(long& costChange)
{
	costChange = 0;
	return false;
}

template<typename MapType> bool ArcCacheST<MapType>::maintainerStep(long& costChange)
{
	bool retval;
	try {
	costChange = 0;
	retval = asyncUpdate(costChange);
	}
	IOPC_LOG_CATCH(IopcException& exc)
		IOPC_LOG_THROW(exc);
	IOPC_LOG_CAUGHT
	if (removeDoomed()) {
		retval = true;
	}
	if (removeZombie()) {
		retval = true;
	}
	return retval;
}

template<typename MapType> long ArcCacheST<MapType>::getSleepTime(const TimeStamp& current, long minValue)
{
	ArcCacheItem* item = NULL;
	long diff = 0;

	if (rActiveList.prev != &rActiveList) {
		diff = rActiveList.prev->timeStamp - current;
		item = rActiveList.prev;
	}
	if (fActiveList.prev != &fActiveList) {
		long d = fActiveList.prev->timeStamp - current;
		if (item == NULL || diff > d) {
			diff = d;
			item = fActiveList.prev;
		}
	}
	if (item == NULL) {
///		return 1800*1000000;	/* something like infinity */
		return 5*1000000;
	}
	if (diff < minValue) {
		return 0;		/* no time to sleep */
	}
	return diff - minValue;
}

}/*namespace iopc*/


#endif /*ARCCACHE_H_*/
