/**
 * @file lruCacheItem.cpp
 *
 * @author Jan Hadrava
 * @date 2004
 * @ingroup iopclibcache
 */

#include "lruCacheItem.h"
#include "lruCacheBase.h"

#include "cache.h"
#include "iopcdb/connection.h"
#include "iopccommon/trace.h"

namespace iopc {

LruCacheItem::LruCacheItem()
: CacheLock(), lock(),
	cache(NULL), strat(), persId(), objectPtr(NULL), state(LruCacheItem::ST_BORN),
	next(this), prev(this), isLocked(false), timeStamp()
{
}

LruCacheItem::LruCacheItem(LruCacheBase& cache, const Strategy& strat, const PersIdentification persId, DatabaseObject* obj, State state)
: CacheLock(), lock(),
	cache(&cache), strat(strat), persId(persId), objectPtr(obj), state(state),
	next(NULL), prev(NULL), isLocked(false), timeStamp()
{
}

LruCacheItem::~LruCacheItem()
{
	IOPC_FUNC_INVOKED;
	IOPC_ASSERT((prev == NULL) == (next == NULL));
	if (prev != NULL) {
		IOPC_ASSERT(next != NULL);
		prev->next = next;
		next->prev = prev;
	}
	try {
		lock.xlock();	/* wait for all locks to be unlocked */
		lock.unlock();
		if (objectPtr != NULL) {
			//objectPtr->FreeAll();
			delete objectPtr;
		}
	} catch (...) {
		IOPC_UNREACHABLE;
	}
	IOPC_FUNC_ENDED;
}

DatabaseObject& LruCacheItem::getObject()
{
	IOPC_FUNC_INVOKED;
	if (objectPtr == NULL) {
		IOPC_LOG_THROW(IopcIncorrectUsageException());
	}
	return *objectPtr;
}

CacheLock& LruCacheItem::addLock(bool rdonly)
{
	IOPC_FUNC_INVOKED;
	return cache->addLock(*this, rdonly);
}

void LruCacheItem::rmLock()
{
	IOPC_FUNC_INVOKED;
	/* this can require some change in cache queues */
	cache->rmLock(*this);
	IOPC_FUNC_ENDED;
}

void LruCacheItem::replaceObject(DatabaseObject* obj)
{
	if (objectPtr != NULL) {
		//objectPtr->FreeAll();
		delete objectPtr;
	}
	objectPtr = obj;
	IOPC_FUNC_ENDED;
}

void LruCacheItem::remove()
{
	IOPC_FUNC_INVOKED;
	IOPC_ASSERT((prev == NULL) == (next == NULL));
	if (prev != NULL) {
		prev->next = next;
		next->prev = prev;
		prev = next = NULL;
	}
	IOPC_FUNC_ENDED;
}

void LruCacheItem::addNext(LruCacheItem& item)
{
	IOPC_FUNC_INVOKED;
	next->prev = &item;
	item.next = next;
	item.prev = this;
	next = &item;
}

void LruCacheItem::addPrev(LruCacheItem &item)
{
	IOPC_FUNC_INVOKED;
	prev->next = &item;
	item.prev = prev;
	item.next = this;
	prev = &item;
}

void LruCacheItem::release()
{
	IOPC_FUNC_INVOKED;
	if (objectPtr != NULL) {
		//objectPtr->FreeAll();
		delete objectPtr;
		objectPtr = NULL;
	}
	IOPC_FUNC_ENDED;
}

}/*namespace iopc*/
