/**
 * @file constCachePtr.h
 *
 * @author Jan Hadrava
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#ifndef CONSTCACHEPTR_H_
#define CONSTCACHEPTR_H_

#include "cache/cacheLock.h"
#include "persIdentification.h"
#include "db/cachedConnection.h"
#include "iopccommon/countedPtr.h"

namespace iopc {

/**
 * @brief A cache pointer that locks the objects with shared read-only locks.
 *
 * The existence of the cache pointer guarantees
 * that the object is loaded into the memory and that the memory address of the object will not change until
 * the instance of the cache pointer is destroyed (unlocked). Such process locks the objects in the cache
 * so they can't be removed unless they are unlocked.
 *
 * @ingroup iopclib
 */
template<typename T> class ConstCachePtr{

public:
	/**
	 * @brief Creates an invalid cache pointer
	 */
	ConstCachePtr();

	/**
	 * @brief Crates a cache pointer to a persistent object
	 *
	 * Requested object is looked up in the cache associated with the @a conn connection
	 * and if not found the object is loaded from the underlying database.
	 *
	 * @param conn A connection with an associated cache in which the object should be locked
	 * @param persId Identity of the object to be locked
	 */
	ConstCachePtr(CachedConnection& conn, const PersIdentification& persId);

	/**
	 * @brief Creates a cache pointer to a transient object owned by the @a DbPtr<T>
	 */
	ConstCachePtr(CountedPtr<T> objectPtr);

	/**
	 * @brief Crates a cache pointer from an already locked object
	 */
	ConstCachePtr(CacheLock& cLock);

	/**
	 * @brief Copy constructor
	 */
	ConstCachePtr(const ConstCachePtr<T>& cptr);

	~ConstCachePtr();

	/**
	 * @brief Copy operator
	 */
	ConstCachePtr<T> &operator=(const ConstCachePtr<T>& src);

	/**
	 * @brief Releases the lock and invalidates the cache pointer
	 */
	void release();

	/**
	 * @brief Determines whether the pointer is valid or not. (Points to a cache-locked instance)
	 */
	inline bool isValid() const;

	/**
	 * @brief Determines whether the pointer is valid or not. (Points to a cache-locked instance)
	 */
	inline operator bool() const;

	/**
	 * @brief Dereferences cache pointer
	 *
	 * @return Counted pointer to the referenced object (points to NULL for invalid cache pointer).
	 */
	inline const CountedPtr<T> operator*() const;

	/**
	 * @brief Dereferences cache pointer
	 *
	 * @return Counted pointer to the referenced object (points to NULL for invalid cache pointer).
	 */
	inline const CountedPtr<T> operator->() const;

	/**
	 * @brief Dereferences cache pointer
	 *
	 * @return Counted pointer to the referenced object (points to NULL for invalid cache pointer).
	 */
	inline const CountedPtr<T> getObjectPtr() const;

protected:
	/**
	 * @brief Counted pointer, for both persistent and transient instances.
	 */
	CountedPtr<T> objectPtr;

	/**
	 * @brief Cache item manipulator (NULL for transient instances).
	 */
	CacheLock* cLock;
};

template<typename T> ConstCachePtr<T>::ConstCachePtr()
{
	cLock = NULL;
	objectPtr = NULL;
}

template<typename T> ConstCachePtr<T>::ConstCachePtr(CachedConnection& conn, const PersIdentification& persId)
{
	cLock = &conn.getCacheLock(persId, true);
	objectPtr = CountedPtr<T>(dynamic_cast<T*>(&cLock->getObject()), false);
}

template<typename T> ConstCachePtr<T>::ConstCachePtr(CountedPtr<T> objectPtr)
{
	cLock = NULL;
	this->objectPtr = objectPtr;
}

template<typename T> ConstCachePtr<T>::ConstCachePtr(const ConstCachePtr<T>& src)
{
	objectPtr = src.objectPtr;
	if (src.cLock == NULL) {	/* transient instance */
		cLock = NULL;
	} else {			/* persistent instance */
		cLock = &src.cLock->addLock(true);
		IOPC_ASSERT(objectPtr == (T*) &cLock->getObject());
	}
}

template<typename T> ConstCachePtr<T>::ConstCachePtr(CacheLock& cLock)
{
	this->cLock = &cLock;
	objectPtr = CountedPtr<T>(dynamic_cast<T*>(&cLock.getObject()), false);
}

template<typename T> ConstCachePtr<T>::~ConstCachePtr()
{
	IOPC_FUNC_INVOKED;
	if (cLock != NULL) {
		cLock->rmLock();
	}
}

template<typename T> ConstCachePtr<T>& ConstCachePtr<T>::operator=(const ConstCachePtr<T>& src)
{
	if (this == &src)
		return *this;
	objectPtr = src.objectPtr;
	if (cLock != NULL) {
		cLock->rmLock();
	}
	if (src.cLock == NULL) {	/* transient instance */
		cLock = NULL;
	} else {			/* persistent instance */
		cLock = &src.cLock->addLock(true);
		IOPC_ASSERT(objectPtr == (T *)&cLock->getObject());
	}
	return *this;
}

template<typename T> void ConstCachePtr<T>::release()
{
	if (cLock != NULL) {
		cLock->rmLock();
		cLock = NULL;
	}
	objectPtr = NULL;
}

template<typename T> inline bool ConstCachePtr<T>::isValid() const
{
	return !objectPtr.isNull();
}

template<typename T> inline ConstCachePtr<T>::operator bool() const
{
	return isValid();
}

template<typename T> const CountedPtr<T> ConstCachePtr<T>::operator*() const
{
	return getObjectPtr();
}

template<typename T> const CountedPtr<T> ConstCachePtr<T>::operator->() const
{
	return getObjectPtr();
}

template<typename T> const CountedPtr<T> ConstCachePtr<T>::getObjectPtr() const
{
	return objectPtr;
}

} /*namespace*/


#endif /*CONSTCACHEPTR_H_*/
