/**
 * @file keyValues.cpp
 *
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#include "keyValues.h"
#include <ostream>
#include <sstream>
#include <iostream>


using namespace std;

namespace iopc {

bool KeyValues::contains(const string& key) const
{
	map<string, EnhancedTypeBase*>::const_iterator it = values.find(key);
	return it != values.end();
}

EnhancedTypeBase& KeyValues::get(const string& key) const
{
	map<string, EnhancedTypeBase*>::const_iterator it = values.find(key);
	if (it == values.end()) {
		IOPC_LOG_THROW(IopcNotFoundException((string) "KeyValues does not contain key " + key));
	}
	return *(it->second);
}

KeyValues::KeyValues(const KeyValues& src)
{
	copyFrom(src);
}

KeyValues::~KeyValues()
{
	clear();
}

KeyValues::KeyValues()
{
}

void KeyValues::clear()
{
	for (map<string, EnhancedTypeBase*>::iterator it = values.begin(); it != values.end(); it++) {
		delete (it->second);
	}
	values.clear();
}

string KeyValues::toString() const
{
	ostringstream os;
	for (map<string, EnhancedTypeBase*>::const_iterator it = values.begin(); it != values.end(); it++) {
		os << it->first << ": " << it->second->toString() << ", ";
	}
	return os.str();
}

void KeyValues::add(std::string key, EnhancedTypeBase& value)
{
	add(key, &value);
}

void KeyValues::add(std::string key, EnhancedTypeBase* value)
{
	EnhancedTypeBase* copy = (EnhancedTypeBase*) value->clone();
	values.insert(std::make_pair<std::string, EnhancedTypeBase*>(key, copy));
}

void KeyValues::addNoCopy(std::string key, EnhancedTypeBase* value)
{
	values.insert(std::make_pair<std::string, EnhancedTypeBase*>(key, value));
}

KeyValues& KeyValues::operator=(const KeyValues& src)
{
	clear();
	copyFrom(src);
	return *this;
}

void KeyValues::copyFrom(const KeyValues& src)
{
	for (map<string, EnhancedTypeBase*>::const_iterator it = src.values.begin(); it != src.values.end(); it++) {
		EnhancedTypeBase* copy = (EnhancedTypeBase*) it->second->clone();
		values.insert(std::make_pair<std::string, EnhancedTypeBase*>(it->first, copy));
	}
}

bool KeyValues::operator<(const KeyValues& src) const
{
	return compare(src) < 0;
}

bool KeyValues::operator==(const KeyValues& src) const
{
	return compare(src) == 0;
}

bool KeyValues::operator!=(const KeyValues& src) const
{
	return compare(src) != 0;
}

int KeyValues::compare(const KeyValues& src) const
{
	if (values.size() < src.values.size()) {
		return -1;
	} else if (values.size() > src.values.size()) {
		return 1;
	}
	if (values.size() == 0) {
		return 0;
	}

	map<string, EnhancedTypeBase*>::const_iterator it = values.begin();
	map<string, EnhancedTypeBase*>::const_iterator it2 = src.values.begin();

	for (; it != values.end() && it2 != src.values.end(); it++, it2++) {
		if (it->first < it2->first) return -1;
		if (it->first > it2->first) return 1;
		EnhancedTypeBase* type1 = it->second;
		EnhancedTypeBase* type2 = it2->second;

		int res = type1->compare(*type2);
		if (res != 0) return res;
	}
	return 0;
}


template<>
void KeyValues::add<const string&>(string key, const string& value) {
	add(key, EString(value));
}
template<>
void KeyValues::add<const wstring&>(string key, const wstring& value) {
	add(key, EWString(value));
}
template<>
void KeyValues::add<const char*>(string key, const char* value) {
	add(key, EString(value));
}
template<>
void KeyValues::add<const wchar_t*>(string key, const wchar_t* value) {
	add(key, EWString(value));
}

template<>
void KeyValues::add<int>(string key, int value) { add(key, ENumber<int>(value)); }
template<>
void KeyValues::add<unsigned int>(string key, unsigned int value) { add(key, ENumber<unsigned int>(value)); }
template<>
void KeyValues::add<char>(string key, char value) { add(key, ENumber<char>(value)); }
template<>
void KeyValues::add<unsigned char>(string key, unsigned char value) { add(key, ENumber<unsigned char>(value)); }
template<>
void KeyValues::add<signed char>(string key, signed char value) { add(key, ENumber<signed char>(value)); }
template<>
void KeyValues::add<short>(string key, short value) { add(key, ENumber<short>(value)); }
template<>
void KeyValues::add<unsigned short>(string key, unsigned short value) { add(key, ENumber<unsigned short>(value)); }
template<>
void KeyValues::add<long>(string key, long value) { add(key, ENumber<long>(value)); }
template<>
void KeyValues::add<unsigned long>(string key, unsigned long value) { add(key, ENumber<unsigned long>(value)); }
template<>
void KeyValues::add<float>(string key, float value) { add(key, ENumber<float>(value)); }
template<>
void KeyValues::add<double>(string key, double value) { add(key, ENumber<double>(value)); }
template<>
void KeyValues::add<long double>(string key, long double value) { add(key, ENumber<long double>(value)); }
template<>
void KeyValues::add<wchar_t>(string key, wchar_t value) { add(key, ENumber<wchar_t>(value)); }
template<>
void KeyValues::add<bool>(string key, bool value) { add(key, ENumber<bool>(value)); }


}
