/**
 * @file query.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#ifndef QUERY_H_
#define QUERY_H_

#include <string>
#include "iopccommon/defs.h"

#define TYPE_DELIMITER "$"

namespace iopc {

class Type;
class Connection;
class Driver;

/**
 * @brief Base query class
 *
 * The class serves as an interface for other @a Quary implementations.
 *
 * The only funcitonality provided by this class is the @a substituteQuery method
 * which substitutes type or attribute names and metadata codes in a query.
 *
 * See @a substituteQuery documentation for the substitution rules.
 *
 * @ingroup iopclib
 */
class DLL_EXTERNAL Query {
public:

	/**
	 * @brief Returns SQL SELECT statement representing the current query
	 *
	 * @param drv Database driver used to generate the statement
	 * @param type Type for which to generate the statement. Determines the columns/attributes that the
	 * SELECT returns.
	 */
	virtual std::string getSql(Driver& drv, const Type& type) const =0;

	const std::string typeDelimiter;

	virtual ~Query() {};

	virtual Query* clone() const =0;

protected:

	/**
	 * @brief Substitutes type, attribute names and metadata codes in a query
	 *
	 * In a query it searches for the TYPE_DELIMITER delimited strings which surround
	 * type, attribute names with optional metadat code attached and replaces them with
	 * type, view names or metadata values.
	 * Format of the delimited strings is:
	 * @code
	 * $(namespace)::TypeName$		// Represents a type (table, view)
	 * $(namespace)::TypeName::AttributeName$    // Represents an attribute (qualified column name)
	 * @endcode
	 *
	 * To specify metadata for which the TypeName or AttributeName is substituted, following format can be used:
	 * @code
	 * $::TypeName[metadata.code]$		// Type is substituted for its metadata value
	 * $::TypeName::AttributeName[metadata.code]$		// Attribute is substituted for its metadata value
	 * $::TypeName[metadata.code]::AttributeName[metadata.code]$	// Both the Type and Attribute are substituted for the metadata specified
	 * @endcode
	 *
	 * If no metadata is used (like in the first example) and if OidObject is queried,
	 * the [db.view.pv] table metadata and [db.viewColumn] attribute metadata is used.
	 * If non-OID object is queried, the [db.table] and [db.column] metadata are used.
	 *
	 * A special [db.query.skip] table metadata can be specified to skip the table or view column qualification:
	 * @code
	 * $::TypeName[db.query.skip]::AttributeName$
	 * // translates only to a column name instead of table(view).columnName
	 * @endcode
	 *
	 * @param query The query which to translate
	 * @param type A @a Type whose instance will be loaded using the query
	 * @param useSV Forces to use simple view instead of polymorfphic views
	 */
	std::string substituteQuery(std::string query, const Type& type, bool useSV = false) const;

	Query() {};

};

}

#endif /*QUERY_H_*/
