/**
 * @file resultBase.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#ifndef RESULTBASE_H_
#define RESULTBASE_H_

#include <string>
#include <vector>
#include "query.h"

namespace iopc {
class CachedConnection;
class DbType;
class Type;

/**
 * @brief Base class for the @a Result<T> template
 *
 * @ingroup iopclib
 */
class DLL_EXTERNAL ResultBase{
protected:
	CachedConnection* conn;
	const Query* query;
	bool isTransient;
	bool preload;
private:
	const Type& type;
	DbType& dbType;

	Cursor* cur;
	KeyValues fetchedKey;
	std::vector<KeyValues> keys;

public:
	ResultBase(CachedConnection* conn, const Query& query, bool preload, const Type& type);
	virtual ~ResultBase();

	/**
	 * @brief Returns the query which was passed to the @a Result<T> constructor
	 */
	inline const Query& getQuery() { return *query; }

	/**
	 * @brief Executes the query and opens the result set for iteration
	 *
	 * Opens a cursor that iterates the query results
	 * If @a preload is specified, all keys (OIDs or KeyValues) are loaded immediately and
	 * the cursor is closed.
	 */
	void open();

	/**
	 * @brief Fetches next result
	 */
	bool fetchNext();

	/**
	 * @brief Fetches previous result
	 */
	bool fetchPrev();

	/**
	 * @brief Closes the @a Result
	 *
	 * If @a preload was not specified, closes the underlying cursor.
	 * Frees all allocated resources.
	 */
	void close();

protected:
	KeyValues getCurrentKey();
	KeyValues getKeyAt(int i);
	inline int getKeyCount() { return keys.size(); }
	virtual bool loadCurrentObject(Cursor& cur) =0;

};

}

#endif /*RESULTBASE_H_*/
