/**
 * @file simpleQuery.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#ifndef SIMPLEQUERY_H_
#define SIMPLEQUERY_H_

#include <string>
#include "iopccommon/defs.h"
#include "query.h"

namespace iopc {


class Type;
class Connection;
class Driver;

/**
 * @brief A query implementation that allows to filter and order persistent instances of a particular type
 *
 * @a SimpleQuery is used to create basic queries which select data only from a single view or table
 * specified by the result type of the query result @a Result<T>. The result can be ordered and filtered
 * by restricting attribute values of the type. @a SimpleQuery is used to retrieve instances of one particular type.
 * It represents only the WHERE and ORDER BY parts of the resulting SQL SELECT statement.
 *
 * @ingroup iopclib
 */
class DLL_EXTERNAL SimpleQuery : public Query {
private:
	std::string query;
	bool all;
	std::string sort;

public:
	/**
	 * @brief A query retrieving all unordered persistent instances of a particular type
	 */
	static SimpleQuery All;

	/**
	 * @param queryStr Query filter which restricts attribute values of the queried type. May be empty.
	 * Example: "::Person::age > 30"
	 * @param sortStr Describes how the results are ordered. May be empty
	 * Example:: "::Person::name ASC"
	 *
	 * Both strings are substituted according to rules described in @a Query::substituteQuery()
	 */
	SimpleQuery(std::string queryStr = "", std::string sortStr = "") : query(queryStr), all(queryStr == ""), sort(sortStr) {};
	SimpleQuery(const SimpleQuery& query) : query(query.query), all(query.all), sort(query.sort) {};

	virtual std::string getSql(Driver& drv, const Type& type) const;

protected:
	virtual Query* clone() const;

};

}

#endif /*SIMPLEQUERY_H_*/
