/**
 * @file dbType.cpp
 *
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#include "dbType.h"
#include "iopccommon/trace.h"
#include "iopcmeta/attributeValue.h"
#include "iopcdb/cursor.h"
#include "dbTypeManager.h"
#include "iopcmeta/enhancedTypeBase.h"
#include "iopccommon/util.h"
#include "iopcmeta/enhancedString.h"
#include "iopcdb/driverManager.h"
//#include "db/sqlStatementsFeature.h"
//#include "../db/objectMappingFeature.h"
#include "iopccommon/globalSettings.h"
#include "keyValues.h"
#include "objectMappingStrategy.h"
#include "orMappingStrategy.h"

#include <sstream>
#include <string>
#include <iostream>
#include <set>
#include <queue>


using namespace std;

namespace iopc {

DbType::DbType()
{
	initialized = false;
	strategy = NULL;
}

void DbType::init(const Type& type)
{
	IOPC_FUNC_INVOKED;
	if (initialized) return;
	IOPC_DEBUGLOG(string("[class]: ") + type.getQualifiedName());
	initialized = true;

	this->type = &type;

	oidManaged = getType().isTypeOf<OidObject>();

	// Check for attribute metadata and find all persistent attributes

	// Set default values for metadata if not set
	if (!getType().hasMeta("db.table")) {
		getType()["db.table"].setStringValue(getType().getName());
	}
	if (!getType().hasMeta("db.view.sv")) {
		getType()["db.view.sv"].setStringValue(getType().getMeta("db.table").getStringValue() + "_SV");
	}
	if (!getType().hasMeta("db.view.pv")) {
		getType()["db.view.pv"].setStringValue(getType().getMeta("db.table").getStringValue() + "_PV");
	}

	const Type::AttributesMap& attributes = getType().getAttributes();
	for (Type::AttributesIterator it = attributes.begin(); it != attributes.end(); it++) {

		if (it->second.hasFeature("db.transient")) {
			continue;
		}
		if (!it->second.hasMeta("db.column")) {
			it->second["db.column"].setStringValue(it->first);
		}
		if (!it->second.hasMeta("db.viewColumn")) {
			it->second["db.viewColumn"].setStringValue(it->second.getMeta("db.column").getStringValue());
			//it->second["db.viewColumn"].setStringValue(getType().getMeta("db.table").getStringValue() + "_" + it->second.getMeta("db.column").getStringValue());
		}

		if (it->second.getType() == TypeDesc<EString>::getType() || it->second.getType() == TypeDesc<EWString>::getType()) {
			if (!it->second.hasMeta("db.type.length")) {
				IOPC_LOG_THROW(IopcMetadataException("db.type.length", it->second.getType().getQualifiedName() + "." + it->first, "Metadata expected for the string attribute"));
			}
		}
		IOPC_DEBUGLOG(string("[persistentAttributes]: ") + it->first);
		persistentAttributes.push_back(&(it->second));

		if (it->second.hasMeta("db.primaryKey")) {
			if (it->second.getType().getDataTypeClass() != Type::IOPC_TYPECLASS_ENHANCED) {
				IOPC_LOG_THROW(IopcMetadataException("db.primaryKey", it->second.getType().getQualifiedName() + "." + it->first, "Attribute marked as primary key must be of an enhanced type"));
			}
			keyAttributes.insert(&(it->second));
		}
	}

	// Mapping
	string tmp = getType().getMeta("db.mapping.type").getStringValue();

	if (tmp == "horizontal") {
		mapping = MAPPING_HORIZONTAL;
		strategy = new ORMappingStrategy(this);
	} else if (tmp == "vertical") {
		mapping = MAPPING_VERTICAL;
		strategy = new ORMappingStrategy(this);
	} else if (tmp == "filtered") {
		mapping =  MAPPING_FILTERED;
		strategy = new ORMappingStrategy(this);
	} else if (tmp == "object") {
		mapping = MAPPING_OBJECT;
		strategy = new ObjectMappingStrategy(this);

		// Default value for db.table
		if (!getType().hasMeta("db.type")) {
			getType()["db.type"].setStringValue(string("t") + getType().getName());
		}
	} else {
		IOPC_LOG_THROW(IopcMetadataException("db.mapping.type", getType().getQualifiedName(), "Incorrect value"));
	}

	queue<const Type*> que;
	que.push(&type);
	do {
		const Type* t = que.front();
		que.pop();
		DbType& dbType = DbTypeManager::getDbTypeWithInit(*t);

		// Stop walking through parents of types with MAPPING_HORIZONTAL mapping
		if (dbType.mapping == MAPPING_HORIZONTAL) continue;

		for(Type::ParentsIterator it = t->getParents().begin(); it != t->getParents().end(); it++) {

			// OidObject will be added after this loop
			if ((*it)->isTypeOf<DatabaseObject>() && (*it)->isNot<OidObject>() &&  (*it)->isNot<DatabaseObject>()) {

				DbType& parentDbType = DbTypeManager::getDbTypeWithInit(*(*it));

				// Filtered or object types don't have associated database table
				if (parentDbType.mapping != MAPPING_FILTERED &&
					parentDbType.mapping != MAPPING_OBJECT) {

					IOPC_DEBUGLOG(string("[allParents]: ") + (*it)->getQualifiedName());
					if (allParentsSet.find(&parentDbType) == allParentsSet.end()) {
						allParents.push_back(&parentDbType);
						allParentsSet.insert(&parentDbType);
					}
				}
				que.push(*it);
			}
		}
	} while (!que.empty());

	DbType& oidObjectDbType = DbTypeManager::getDbTypeWithInit(TypeDesc<OidObject>::getType());
	if (oidManaged && type.isNot<OidObject>() && oidObjectDbType.getMappingType() != MAPPING_OBJECT) {
		IOPC_DEBUGLOG(string("[allParents]: ") + TypeDesc<OidObject>::getType().getQualifiedName());
		allParents.push_back(&oidObjectDbType);
		allParentsSet.insert(&oidObjectDbType);
	}

	/*if (allParents.size() == 0) {
		mapping = MAPPING_HORIZONTAL;
	}*/

	if (!allParents.empty() && !keyAttributes.empty()) {
		IOPC_LOG_THROW(IopcMetadataException("db.primaryKey", getType().getQualifiedName() + "." + (*keyAttributes.begin())->getName(), "Can't specify db.primaryKey. Primary keys are inherited from the top parent. (OidObject even does not need to specify the primary key)"));
	}


	// Check for type metadata
	if (mapping != MAPPING_FILTERED && !getType().hasMeta("db.table")) {
		IOPC_LOG_THROW(IopcMetadataException("db.table",  getType().getQualifiedName(), "No such metadata found"));
	}

	strategy->init();


	IOPC_DEBUGLOG(string("[end class]: ") + type.getQualifiedName());
	IOPC_FUNC_ENDED;
}


void DbType::postInit()
{
	IOPC_FUNC_INVOKED;
	for (Type::AllParentsIterator it = getType().getAllParents().begin(); it != getType().getAllParents().end(); it++) {
		if (!(*it)->isTypeOf<DatabaseObject>() || (*it)->is<DatabaseObject>()) continue;
		DbType& t = DbTypeManager::getDbType(*(*it));

		inheritedAttributes.insert(inheritedAttributes.end(), t.persistentAttributes.begin(), t.persistentAttributes.end());


		if (!t.keyAttributes.empty() && keyAttributes.empty()) {
			// Inherit keys from parent
			keyAttributes = t.keyAttributes;
		}
		// TODO: kontrola pokud parent i child maji deklarovane key attributy, pak chybu ze maji byt jen jednou v parentovi
	}
	if (keyAttributes.empty() && !oidManaged && !getType().hasFeature("db.transient")) {
		IOPC_LOG_THROW(IopcMetadataException("db.primaryKey", getType().getQualifiedName(), "At least one-column primary key expected in the top-level non-oid based type."));
	}

	allPersistentAttributes.insert(allPersistentAttributes.end(), persistentAttributes.begin(), persistentAttributes.end());
	allPersistentAttributes.insert(allPersistentAttributes.end(), inheritedAttributes.begin(), inheritedAttributes.end());

#ifdef IOPC_DEBUG
		IOPC_DEBUGLOG(string("[class]: ") + getType().getQualifiedName());
		for (DbType::PersistentAttributesIterator it = inheritedAttributes.begin(); it != inheritedAttributes.end(); it++) {
			IOPC_DEBUGLOG(string("[inheritedAttributes]: ") + (*it)->getName());
		}
		for (DbType::KeyAttributesIterator it = keyAttributes.begin(); it != keyAttributes.end(); it++) {
			IOPC_DEBUGLOG(string("[keyAttributes]: ") + (*it)->getName());
		}
		for (DbType::PersistentAttributesIterator it = allPersistentAttributes.begin(); it != allPersistentAttributes.end(); it++) {
			IOPC_DEBUGLOG(string("[allPersistentAttributes]: ") + (*it)->getName());
		}
#endif

	strategy->postInit();
}



/*DbType& DbType::operator=(const DbType& src)
{
	type = src.type;
	init();
	//insertStatement = src.insertStatement;
	//persistentAttributes = src.persistentAttributes;
	return *this;
}*/


DbType::~DbType()
{
	if (strategy != NULL) {
		delete strategy;
	}
}


void DbType::loadOne(Connection& conn, DatabaseObject* obj, const KeyValues& keyVals, bool wait, bool lock)
{
	strategy->loadOne(conn, obj, keyVals, wait, lock);
}

DatabaseObject* DbType::loadDatabaseObject(Connection& conn, const KeyValues& keyVals, bool wait, bool lock)
{
	IOPC_FUNC_INVOKED;

	DatabaseObject* obj = dynamic_cast<DatabaseObject*>(type->newObjectInstance());
	try {
		loadOne(conn, obj, keyVals, wait, lock);

		// and all his ancestors
		for (AllParentsIterator it = allParents.begin(); it != allParents.end(); it++) {
			(*it)->loadOne(conn, obj, keyVals, wait, lock);
		}

		return obj;
	} catch (...) {
		delete obj;
		throw ;
	}
}

DatabaseObject* DbType::loadDatabaseObject(Cursor& cur)
{
	IOPC_FUNC_INVOKED;
	DatabaseObject* obj = dynamic_cast<DatabaseObject*>(type->newObjectInstance());
	try {
		//loadOne(conn, obj, keyVals, wait, lock);

		// and all his ancestors
		/*for (AllParentsIterator it = allParents.begin(); it != allParents.end(); it++) {
			(*it)->loadOne(conn, obj, keyVals, wait, lock);
		}*/


		int i = 1;
		for (PersistentAttributesIterator it = allPersistentAttributes.begin(); it != allPersistentAttributes.end(); it++) {
			const AttributeValue& attrVal = obj->getAttributeValue((*it)->getName());
			cur.addOutParam(i, attrVal.getValue(), attrVal.getAttribute().getType(), (MetadataHolder&) attrVal.getAttribute());
			i++;
		}

		//cur->execute();
		if (!cur.fetchNext()) {
			delete obj;
			return NULL;
		}
		//cur->close();
		//conn.returnCursor(cur);


		return obj;
	} catch (...) {
		delete obj;
		throw ;
	}
}

OidObject* DbType::loadOidObject(Connection& conn, const OidType& oid, bool wait, bool lock)
{
	IOPC_FUNC_INVOKED;

	if (lock) {
		// we have to load this object as it was a database object to be able to lock its rows.
		KeyValues keyVals;
		keyVals.add("oid", oid);
		DatabaseObject* obj = loadDatabaseObject(conn, keyVals, wait, lock);
		return dynamic_cast<OidObject*>(obj);
	}

	return strategy->viewLoadAll(conn, oid, wait, lock);
}

void DbType::updateOne(Connection& conn, DatabaseObject& obj)
{
	strategy->updateOne(conn, obj);
}

void DbType::updateObject(Connection& conn, DatabaseObject& obj)
{
	IOPC_FUNC_INVOKED;
	// Update this object
	updateOne(conn, obj);

	// and all his ancestors
	for (AllParentsIterator it = allParents.begin(); it != allParents.end(); it++) {
		(*it)->updateOne(conn, obj);
	}

	IOPC_FUNC_ENDED;
}

void DbType::deleteOne(Connection& conn, const OidType& oid)
{
	strategy->deleteOne(conn, oid);
}

void DbType::deleteOne(Connection& conn, const KeyValues& keyVals)
{
	strategy->deleteOne(conn, keyVals);
}

void DbType::deleteOidObject(Connection& conn, const OidType& oid)
{
	IOPC_FUNC_INVOKED;
	// Update this object
	deleteOne(conn, oid);

	// and all his ancestors
	for (AllParentsIterator it = allParents.begin(); it != allParents.end(); it++) {
		(*it)->deleteOne(conn, oid);
	}
	IOPC_FUNC_ENDED;
}

void DbType::deleteDatabaseObject(Connection& conn, const KeyValues& keyVals)
{
	IOPC_FUNC_INVOKED;
	// Update this object
	deleteOne(conn, keyVals);

	// and all his ancestors
	for (AllParentsIterator it = allParents.begin(); it != allParents.end(); it++) {
		(*it)->deleteOne(conn, keyVals);
	}
	IOPC_FUNC_ENDED;
}

void DbType::insertOne(Connection& conn, DatabaseObject& obj)
{
	strategy->insertOne(conn, obj);
}

void DbType::insertObject(Connection& conn, DatabaseObject& obj)
{
	IOPC_FUNC_INVOKED;
	/*if (oidManaged) {
		obj.oid = getNextOid(conn);
	}*/

	// All ancestors
	for (AllParentsReverseIterator it = allParents.rbegin(); it != allParents.rend(); it++) {
		(*it)->insertOne(conn, obj);
	}

	// Insert this object
	insertOne(conn, obj);

	IOPC_FUNC_ENDED;
}

KeyValues DbType::getKeyValuesForObject(const DatabaseObject& obj) const
{
	IOPC_FUNC_INVOKED;
	KeyValues r;

	for (KeyAttributesIterator it = keyAttributes.begin(); it != keyAttributes.end(); it++)
	{
		string attrName = (*it)->getName();
		r.add(attrName, (EnhancedTypeBase*) obj.getAttributeValue(attrName).getValue());
	}

	return r;
}

ObjectMappingStrategy& DbType::getObjectMappingStrategy() const
{
	if (getMappingType() != MAPPING_OBJECT) {
		IOPC_LOG_THROW(IopcIncorrectUsageException("Can't return ObjectMappingStrategy for class with db.mapping.type != 'object'"));
	}
	return (ObjectMappingStrategy&) getMappingStrategy();
}

ORMappingStrategy& DbType::getORMappingStrategy() const
{
	if (getMappingType() == MAPPING_OBJECT) {
		IOPC_LOG_THROW(IopcIncorrectUsageException("Can't return ORMappingStrategy for class with db.mapping.type == 'object'"));
	}
	return (ORMappingStrategy&) getMappingStrategy();
}


}
