/**
 * @file dbTypeManager.h
 *
 * @author Petr Cermak
 *
 * @ingroup iopclib
 */

#ifndef DBTYPEMANAGER_H_
#define DBTYPEMANAGER_H_

#include <map>
#include "iopcmeta/type.h"
#include "dbType.h"
#include "iopccommon/lock.h"
#include "iopccommon/rwxLock.h"


namespace iopc {

class Driver;

/**
 * @brief This singleton class manages the @a DbType instances
 *
 * Initializes them, allows to look them up by @a Type associated to them.
 * Provides O/R mapping services which are not @a Type specific.
 *
 * @ingroup iopclib
 */
class DbTypeManager {
public:

	typedef std::map<const Type*, DbType> DbTypesMap;
	typedef DbTypesMap::const_iterator DbTypesIterator;

	/**
	 * @brief Looks up a @a DbType instance by the @a Type it represents
	 */
	static inline DbType& getDbType(const Type& t) {
		return getInstance().getDbTypeInst(t);
	}

	/**
	 * @brief Looks up a @a DbType instance by the @a Type it represents and initializes it
	 */
	static inline DbType& getDbTypeWithInit(const Type& t) {
		DbType& dbType = getInstance().getDbTypeInst(t);
		dbType.init(t);
		return dbType;
	}

	/**
	 * @brief Returns a map of @a DbType instances managed by this singleton
	 */
	static inline DbTypesMap& getDbTypesMap() {
		return getInstance().getDbTypesMapInst();
	}


	/**
	 * @brief Initializes the @a DbTypeManager
	 *
	 * Initializes all registered @a DbType instances
	 * Invoked from @a Iopc::init()
	 *
	 * @see Iopc::init()
	 */
	static inline void init() {
		getInstance().initInst();
	}

	/**
	 * @brief Returns the singleton instance
	 */
	static DbTypeManager& getInstance();

	/**
	 * @brief Returns an actual type of a persistent object identified by an OID
	 *
	 * @param conn Connection used to send the query to the OID - Classname catalogue
	 * @param oid OID of the object whose type will be retrieved
	 */
	static inline const Type& getType(Connection& conn, const OidType oid) {
		return getInstance().getTypeInst(conn, oid);
	}

	/**
	 * @brief Generates a new OID
	 *
	 * @param conn Connection used to generate new OID
	 */
	static inline const OidType getNextOid(Connection& conn) {
		return getInstance().getNextOidInst(conn);
	}

	/**
	 * @brief Generates a new SerialID
	 *
	 * @param conn Connection used to generate new SerialID
	 */
	static inline const SerialIdType getNextSerialId(Connection& conn) {
		return getInstance().getNextSerialIdInst(conn);
	}

	/**
	 * @brief Returns a serialId timestamp for specified OID
	 *
	 * @param conn Connection used to send the query to the OID - ClassName catalogue
	 * @param oid OID for which to find the current timestamp
	 */
	static inline const SerialIdType getSerialId(Connection& conn, const OidType oid) {
		return getInstance().getSerialIdInst(conn, oid);
	}


	inline DbTypesMap& getDbTypesMapInst() {
		return dbTypes;
	}
	void initInst();
	DbType& getDbTypeInst(const Type& t);

	const Type& getTypeInst(Connection& conn, const OidType oid);
	OidType getNextOidInst(Connection& conn);
	SerialIdType getNextSerialIdInst(Connection& conn);
	SerialIdType getSerialIdInst(Connection& conn, const OidType oid);



	typedef std::vector<Driver*> DriversList;
	typedef DriversList::const_iterator DriversIterator;

	static inline const DriversList& getORDrivers() { return getInstance().getORDriversInst(); }
	inline const DriversList& getORDriversInst() const { return driversList;}

private:


	DbTypesMap dbTypes;


	DbTypeManager();
	~DbTypeManager();

	DriversList driversList;

};

}


#endif /*DbTypeManager_H_*/
