/**
 * @file object.h
 *
 * @ingroup iopcmeta
 *
 * @author Petr Cermak
 */
#ifndef OBJECT_H_
#define OBJECT_H_

#include "iopccommon/defsPlatform.h"
#include "iopccommon/trace.h"
#include "type.h"
#include "attributeValue.h"
#include "iopccommon/lock.h"
#include <map>

namespace iopc {

class Type;

/**
 * @brief Ancestor for all reflection-capable classes
 *
 * The @c iopcsp utility processes classes that are descendants of @a Object only. Other classes
 * can't be introspected using reflection features provided by this library.
 *
 * @a Object provides infrastructure for dirty status tracking and a list of @a AttributeValue objects
 * that allow developers to modify object attribute values using the reflection.
 *
 * @ingroup piopcmeta
 */
class DLL_EXTERNAL Object
{
public:
	typedef std::map<std::string, AttributeValue> AttributeValuesMap;
	typedef AttributeValuesMap::const_iterator AttributeValuesIterator;

private:
	mutable bool metaInitialized;
	bool dirtyInitialized;
	bool dirty;
	mutable AttributeValuesMap attributes;
	mutable const Type* type;
	mutable Lock lock;

public:
	Object();

	/**
	 * @brief Returns string representation of this object
	 */
	virtual std::string toString() const;

	/**
	 * @brief Returns memory cost of this object.
	 *
	 * Memory cost is computed as a sum over Type::getSizeOf(attribute value).
	 */
	virtual int getMemCost() const;

	/**
	 * @brief Checks whether the object was modified since it was created or its dirty flag was reset.
	 */
	virtual bool isDirty();

	/**
	 * @brief Resets the dirty flag of this object
	 */
	virtual void resetDirty();

	/**
	 * @brief Sets the dirty flag to a specified value
	 */
	virtual void setDirty(bool dirty = true);
	/** This function is called whenever persistent attributes are modified. */
/*	inline virtual void setDirty(bool dirty = true) {
		IOPC_DEBUGLOG("Object::setDirty");
		this->dirty = dirty;
	}*/

	/**
	 * @brief Compares this object with another object
	 *
	 * This implementation compares memory addresses of the involved objects.
	 *
	 * @param o Object to compare with
	 * @returns 0 - equal, -1 this < o, 1 this > o
	 */
	virtual int compare(const Object& o) const;

	inline bool operator<(const Object& o) const {
		return (compare(o) == -1);
	}
	inline bool operator==(const Object& o) const {
		return (compare(o) == 0);
	}
	inline bool operator!=(const Object& o) const {
		return (compare(o) == 1);
	}

public:

	/**
	 * @brief Returns a Type Description representing the type of the current object
	 */
	virtual const Type& getType() const;

	/**
	 * @brief Returns an attributre value of a specified attribute
	 *
	 * @see AttributeValue
	 */
	const AttributeValue& getAttributeValue(const std::string& name) const;

	/**
	 * @brief Returns a list of attribute values of attributes from this object
	 *
	 * @see AttributeValue
	 */
	inline const AttributeValuesMap& getAttributeValues() const { return attributes; }


	/*inline void setTransient() {
		IOPC_DEBUGLOG("Object::setTransient");
	};

	inline void freeAll() throw () {
		IOPC_DEBUGLOG("Object::freeAll");
	};

	inline void initAll() throw () {
		IOPC_DEBUGLOG("Object::initAll");
	};*/

	/**
	 * @brief Returns a copy of this object.
	 *
	 * @a Object descendants may implement this method in various ways.
	 * Default implementation returns only an empty @a Object instance.
	 */
	virtual Object* clone() const;


	virtual ~Object() {};

protected:
	virtual const Type& getTypeInternal() const;

	void initMeta() const;
	void initDirty();
};

}

#endif /*OBJECT_H_*/
