/**
 * @file textFileMetadataLoader.h
 *
 * @ingroup iopcmeta
 *
 * @author Petr Cermak
 */

#ifndef TEXTFILEMETADATALOADER_H_
#define TEXTFILEMETADATALOADER_H_

#include "metadataInitializer.h"
#include <string>

namespace iopc {

class MetadataHolder;

/**
 * @brief A @a MetadataInitializer implementation which loads class metaddta
 * and global settings from a configuration file.
 *
 * The format of the configuration file is described by the following example:
 * @code
 * # global level
 * defaults[db.mapping.type];string;object
 * # type level
 * ::Person[db.type];string;Person_type
 * # attribute level
 * ::Person::name[db.type.length];int;20
 * # global settings
 * globals[db.oracle.workaround.10gOO];bool;true
 * @endcode
 *
 * @a TextFileMetadataLoader skips all empty lines or lines starting with the '#' character.
 * The file is case-sensitive. Lines describing the metadata contain three semicolon delimited values.
 * First value specifies the location and code of the metadata to be set.
 * It uses the same format as metadata in queries. Second column denotes the data type of the metadata value.
 * Only @c string, @c int and @c bool values are allowed. Last column contains the metadata value.
 * Boolean values can be either @c true or @c false.
 *
 * @ingroup iopcmeta
 */
class TextFileMetadataLoader : public MetadataInitializer {
private:
	const std::string file;

public:

	/**
	 * @param file Name (or path) of the configuration file
	 */
	TextFileMetadataLoader(std::string file) : file(file) {};
	virtual ~TextFileMetadataLoader() {};

	virtual void initializeMetadata();
private:
	void setMeta(const MetadataHolder& meta, std::string metadata, std::string type, std::string value);
};

}

#endif /* TEXTFILEMETADATALOADER_H_ */
