/**
 * @file type.h
 *
 * @ingroup iopcmeta
 *
 * @author Petr Cermak
 */


#ifndef TYPE_H_
#define TYPE_H_

#include "iopccommon/defsPlatform.h"
#include "object.h"
#include "attribute.h"
#include "attributeValue.h"
#include "iopccommon/metadataHolder.h"
//#include "meta/typeDesc.h"
#include <string>
#include <map>
#include <vector>
#include <set>

namespace iopc {

class IopcCommon;
class Object;

#define NAMESPACE_DELIMITER "::"

template<typename T> class TypeDesc;


/**
 * @brief Instances of this class descendants - the @e Type @e Descriptions - represent reflection-capable classes and selected C++ or STL types.
 *
 * There is only one Type Description for a reflection-capable class or a basic type.
 * This interface provides methods for the type structure inspection.
 *
 * Implementations of this abstract class - specialisations of the @a TypeDesc<T> template class
 * are either defined in the library or generated using @c iopcsp.
 *
 * The following example illustrates the type inspection features this class provides.
 * @code
 * // Type Descriptions can be obtained either from the TypeManager:
 * const Type& t = TypeManager::getType("::ClassX");
 * // or statically:
 * const Type& t = TypeDesc<Person>::getType();
 *
 * // The type parents can be iterated:
 * const Type::ParentsList& parents = t.getParents();
 * for(Type::ParentsIterator it = parents.begin();	it != parents.end(); it++) {
 *   cout << (*it)->getQualifiedName() << endl;
 * }
 *
 * // Children:
 * const Type::ChildrenList& children = t.getChildren();
 * for(Type::ChildrenIterator it = children.begin(); it != children.end(); it++) {
 *   cout << (*it)->getQualifiedName() << endl;
 * }
 *
 * // Attributes:
 * const Type::AttributesMap& attributes = t.getAttributes();
 * for(Type::AttributesIterator it = attributes.begin(); it != attributes.end(); it++) {
 *   const Attribute& attr = it->second;
 *	 switch (attr.getVisibility()) {
 *		case Attribute::VISIBILITY_PUBLIC: cout << "public "; break;
 *		case Attribute::VISIBILITY_PRIVATE: cout << "private "; break;
 *		case Attribute::VISIBILITY_PROTECTED: cout << "protected ";break;
 *	 }
 *	 cout << it->second.getType().getQualifiedName() 	<< " " << it->first << endl;
 * }
 *
 * // Type Descriptions can be compared:
 * const Type& t1 = TypeDesc<Person>::getType();
 * const Type& t2 = TypeDesc<Student>::getType();
 * if (t1 == t2) { cout << "t1 is t2" << endl; }
 * if (t1 != t2) { cout << "t1 is not t2" << endl; }
 * if (t1.isTypeOf(t2)) { cout << "t1 is t2 or t1 is descendant of t2" << endl; }
 * if (t2.isTypeOf(t1)) { cout << "t1 is t2 or t2 is descendant of t1" << endl; }
 * if (t1.is<Student>()) { cout << "t1 is of type Student" << endl; }
 * if (t2.is<Student>()) { cout << "t2 is of type Student" << endl; }
 * if (t1.isTypeOf<Person>()) { cout << "t1 is of type Person or its descendant" << endl; }
 * if (t2.isTypeOf<Person>()) { cout << "t2 is of type Person or its descendant" << endl; }
 * @endcode
 *
 * @ingroup iopcmeta
 */
class DLL_EXTERNAL Type : public MetadataHolder {

	friend class TypeManager;

public:

	/**
	 * @brief Returns name of the type.
	 */
	virtual const std::string getName() const = 0;

	/**
	 * @brief Returns name of the namespace in which this type is defined
	 *
	 * Returns empty string if the type does not have any namespace.
	 */
	virtual const std::string getNamespace() const = 0;

	/**
	 * @brief Returns qualified name of the type
	 *
	 * @returns getNamespace() + NAMESPACE_DELIMITER + getName()
	 */
	inline const std::string getQualifiedName() const {
		return getNamespace() + NAMESPACE_DELIMITER + getName();
	}

	/**
	 * @brief Returns memory size of object of this type
	 *
	 * In some cases, the @a object parameter is ignored as it is not needed
	 * to determine the memory size. (Types of non-variable length such as numeric types).
	 * For some variable length types such as std::string or EString it doesn't return
	 * sizeof(std::string), sizeof(EString), but memory size needed to store the strings
	 * as standard, zero-terminated C-strings.
	 *
	 * @returns Memory size needed to store the object this type represents.
	 */
	virtual int getSizeOf(const void* object) const =0;

	/**
	 * @brief Determines whether the type is C++ built-in or STL data type
	 */
	virtual bool isBuiltIn() const = 0;

	/**
	 * @brief Allocates a new instance of the data type this @a Type represents
	 *
	 * The instance is allocated using the @a new operator.
	 * To free the instance, use the corresponding @a deleteInstance(void* instance) method.
	 */
	virtual void* newInstance() const = 0;

	/**
	 * @brief Allocates a new instance of the reflection-capable class this @a Type represents
	 *
	 * The instance is allocated using the @a new operator.
	 * To free the instance, you can use standard @a delete operator.
	 *
	 * @returns Other types than @a Object descendants return NULL.
	 */
	virtual Object* newObjectInstance() const =0;

	/**
	 * @brief Frees the instance allocated by the @a newInstance method
	 */
	virtual void deleteInstance(void* instance) const = 0;

	/**
	 * @brief Types can be classified into one of the following categories
	 */
	enum DataTypeClassEnum {
		IOPC_TYPECLASS_SIMPLE = 0,	/**< Built-in numeric types */
		IOPC_TYPECLASS_STRING,		/**< String types - std::string and std::wstring */
		IOPC_TYPECLASS_COMPLEX,		/**< Classes or structures. Descendants of @a Object (except enhanced data types) */
		IOPC_TYPECLASS_ENHANCED		/**< Enhanced data types. Descendants of @a EnhancedTypeBase */
	};

	/**
	 * @brief Returns the data type class for the current type
	 */
	virtual DataTypeClassEnum getDataTypeClass() const = 0;

	/**
	 * @brief Enumerates types defined in @c iopcmeta
	 */
	enum DataTypeEnum {
	  IOPC_TYPE_INVALID = 0, //!< not a valid type

	  IOPC_TYPE_BOOL,        //!< bool

	  IOPC_TYPE_CHAR,        //!< char (signed or unsigned - depends on platform and compiler settings)
	  IOPC_TYPE_UCHAR,       //!< unsigned char
	  IOPC_TYPE_SCHAR,       //!< signed char
	  IOPC_TYPE_SHORT,       //!< short (signed)
	  IOPC_TYPE_USHORT,      //!< unsigned short
	  IOPC_TYPE_INT,         //!< int (signed)
	  IOPC_TYPE_UINT,        //!< unsigned
	  IOPC_TYPE_LONG,        //!< long (signed)
	  IOPC_TYPE_ULONG,       //!< unsigned long

	  IOPC_TYPE_FLOAT,       //!< float
	  IOPC_TYPE_DOUBLE,      //!< double
	  IOPC_TYPE_LONGDOUBLE,  //!< long double

	  IOPC_TYPE_WCHART,	     //!< wchar_t

	  IOPC_TYPE_STRING,      //!< char * (0-terminated)
	  IOPC_TYPE_WSTRING,     //!< wchar_t * (00-terminated)

	  IOPC_TYPE_REF,         //!< reference to some other persistent object

	  IOPC_TYPE_COMPLEX,     //!< Classes or structures. Descendants of @a Object (except enhanced data types)

  	  IOPC_TYPE_EINT,        //!< Enhanced int
  	  IOPC_TYPE_EUINT,       //!< Enhanced unsigned int
  	  IOPC_TYPE_ECHAR,       //!< Enhanced char
  	  IOPC_TYPE_EUCHAR,      //!< Enhanced unsigned char
  	  IOPC_TYPE_ESCHAR,      //!< Enhanced signed char
  	  IOPC_TYPE_ESHORT,      //!< Enhanced short int
  	  IOPC_TYPE_EUSHORT,     //!< Enhanced unsigned short int
  	  IOPC_TYPE_ELONG,       //!< Enhanced long
  	  IOPC_TYPE_EULONG,      //!< Enhanced unsigned long
  	  IOPC_TYPE_EFLOAT,      //!< Enhanced float
  	  IOPC_TYPE_EDOUBLE,     //!< Enhanced double
  	  IOPC_TYPE_ELDOUBLE,    //!< Enhanced long double
  	  IOPC_TYPE_EWCHART,     //!< Enhanced wchar_t
  	  IOPC_TYPE_EBOOL,       //!< Enhanced bool

  	  IOPC_TYPE_ESTRING,     //!< Enhanced string
  	  IOPC_TYPE_EWSTRING     //!< Enhanced wstring
	};

	/**
	 * @brief Returns the @a DataTypeEnum value of the current type
	 */
	virtual DataTypeEnum getDataType() const =0;

	typedef std::vector<const Type*> ParentsList;
	typedef std::vector<const Type*>::const_iterator ParentsIterator;
	typedef std::vector<const Type*> ChildrenList;
	typedef std::vector<const Type*>::const_iterator ChildrenIterator;
	typedef std::set<const Type*> AllParentsList;
	typedef std::set<const Type*>::const_iterator AllParentsIterator;
	typedef std::set<const Type*> AllChildrenList;
	typedef std::set<const Type*>::const_iterator AllChildrenIterator;
	/**
	 * Keys are names of the attributes stored as values
	 */
	typedef std::map<std::string, Attribute> AttributesMap;
	typedef AttributesMap::const_iterator AttributesIterator;
	/**
	 * Keys are names of the attributes stored as values
	 */
	typedef std::map<std::string, const Attribute*> InheritedAttributesMap;
	typedef InheritedAttributesMap::const_iterator InheritedAttributesIterator;

	/**
	 * @brief Used by @a Object
	 *
	 * Used to fill a map of AttributeValues by the type description
	 *
	 * @param[in] o Object containing the attributes
	 * @param[out] out Map into which the attribute values are inserted
	 */
	virtual void fillAttributeValues(const Object& o, std::map<std::string, AttributeValue>& out) const =0;

	/**
	 * @brief Returns a list of direct ancestors of this type
	 */
	inline const ParentsList& getParents() const { return parents; };

	/**
	 * @brief Returns a list of direct descendants of this type
	 */
	inline const ChildrenList& getChildren() const { return children; };

	/**
	 * @brief Returns a list of direct and indirect descendants of this type
	 */
	inline const AllChildrenList& getAllChildren() const { return allChildren; };
	/**
	 * @brief Returns a list of direct and indirect ancestors of this type
	 */
	inline const AllParentsList& getAllParents() const { return allParents; }

	/**
	 * @brief Returns a map of @a Attribute objects representing attributes defined in this type
	 */
	inline const AttributesMap& getAttributes() const { return attributes; };

	/**
	 * @brief Returns a map of @a Attribute objects representing attributes defined in this type
	 * and attributes inherited from its ancestors.
	 */
	inline const InheritedAttributesMap& getInheritedAttributes() const { return inheritedAttributes; }

	/**
	 * @brief Looks up an @a Attribute by its name. Uses the list of attributes defined in this type.
	 *
	 * @param name A name of attribute to look for
	 *
	 * @throws IopcIncorrectUsageException If no attribute of the specified name is found
	 */
	const Attribute& getAttribute(const std::string& name) const;

	/**
	 * @brief Determines whether this type contains an @a Attribute of a specified name
	 *
	 * @param name A name of attribute to look for
	 */
	bool hasAttribute(const std::string& name) const;

	/**
	 * @brief Looks up an @a Attribute by its name. Uses the list of attributes defined in this type
	 * or in its ancestors.
	 *
	 * @param name A name of attribute to look for
	 *
	 * @throws IopcIncorrectUsageException If no attribute of the specified name is found
	 */
	const Attribute& getAttributeAll(const std::string& name) const;

	/**
	 * @brief Determines whether this type or its ancestors contain an @a Attribute of a specified name
	 *
	 * @param name A name of attribute to look for
	 */
	bool hasAttributeAll(const std::string& name) const;

	/**
	 * @brief Assumes that this type has only one parent and returns it
	 *
	 * @throws IopcIncorrectUsageException If this type has no parent class
	 */
	const Type& getFirstParent() const;

	/**
	 * @brief Determines whether this type is descendant of or identitcal to a @a Type
	 *
	 * @param type The type to test against
	 */
	bool isTypeOf(const Type& type) const;

	/**
	 * @brief  Determines whether this type is descendant of or identitcal to a type
	 *
	 * @tparam T The type to test against
	 */
	template<typename T> bool isTypeOf() const {
		return isTypeOf(TypeDesc<T>::getType());
	}

	/**
	 * @brief  Determines whether this type is identitcal to a specified type
	 *
	 * @tparam T The type to test against
	 */
	template<typename T> bool is() const {
		return *this == TypeDesc<T>::getType();
	}

	/**
	 * @brief  Determines whether this type differs from a specified type
	 *
	 * @tparam T The type to test against
	 */
	template<typename T> bool isNot() const {
		return *this != TypeDesc<T>::getType();
	}


public:
	inline bool operator!=(const Type& src) const { return this != &src; }
	inline bool operator==(const Type& src) const { return this == &src; }
	inline bool operator<(const Type& src) const { return this < &src; }

protected:

	Type();
	virtual ~Type();

	virtual void iopcInit() {};

	AttributesMap attributes;
	InheritedAttributesMap inheritedAttributes;
	std::vector<const std::type_info*> parentsTypeInfo;
	ParentsList parents;
	ChildrenList children;
	AllChildrenList allChildren;
	AllParentsList allParents;

	/**
	 * Cached set of all (even indirect) parents.
	 */

private:
	void generateAllChildren();
	void generateAllParents();
	void generateInheritedAttributes();

	/**
	 * Called via TypeManager from a friend class IopcMeta::init()
	 */
	void init();
	void postInit();


	// TODO odstranit - pro debug
	std::string typeName;
};

}

#endif /*TYPE_H_*/
