/**
 * @file typeManager.cpp
 *
 * @ingroup iopcmeta
 *
 * @author Petr Cermak
 */

#include "typeManager.h"
#include <iostream>
#include "iopccommon/exceptionBase.h"
#include "enhancedTypeBase.h"
#include "iopccommon/util.h"

using namespace std;

namespace iopc {


TypeManager::TypeManager()
{
}

TypeManager::~TypeManager()
{
	//types.clear();
}

TypeManager& TypeManager::getInstance()
{
	static TypeManager instance;
	return instance;
}


void TypeManager::registerTypeInst(const Type* type, const std::string& typeName, const type_info* ti)
{
	IOPC_FUNC_INVOKED;
	//std::cout << typeName << std::endl;
	lock.lock();
	types[typeName] = type;
	typeInfoToType[ti] = type;
	lock.unlock();
}

void TypeManager::registerTypeInst(const Type* type, const type_info* ti)
{
	IOPC_FUNC_INVOKED;
	registerTypeInst(type, type->getQualifiedName(), ti);
}

const Type& TypeManager::getTypeInst(const std::string& typeName)
{
	lock.lock();
	const Type* type = types[typeName];
	if (type == NULL) {
		lock.unlock();
		IOPC_LOG_THROW(IopcNotFoundException(string("Could not find type: ") + typeName));
	}
	lock.unlock();
	return *type;
}

bool TypeManager::hasTypeInst(const std::string& typeName)
{
	lock.lock();
	map<string, const Type*>::const_iterator it = types.find(typeName);
	if (it != types.end()) {
		lock.unlock();
		return true;
	} else {
		lock.unlock();
		return false;
	}
}

const Type& TypeManager::getTypeByTypeInfoInst(const type_info* ti)
{
	lock.lock();
	const Type* type = typeInfoToType[ti];
	if (type == NULL) {
		lock.unlock();
		IOPC_LOG_THROW(IopcNotFoundException(string("Could not find type with type_info of name: ") + ti->name()));
	}
	lock.unlock();
	return *type;
}

/**
 * WARNING: iterating the list of types is not thread-safe against registering a new type.
 */
/*map<string, const Type*>::iterator TypeManager::getTypesIteratorInst()
{
	return types.begin();
}*/

const std::map<std::string, const Type*>& TypeManager::getTypesMapInst() const
{
	return types;
}

void TypeManager::initInst()
{
	lock.lock();
	for(map<string, const Type*>::const_iterator it = types.begin(); it != types.end(); it++) {
		const_cast<Type*>((*it).second)->init();
	}
	for(map<string, const Type*>::const_iterator it = types.begin(); it != types.end(); it++) {
		const_cast<Type*>((*it).second)->postInit();
	}
	lock.unlock();

	EnhancedTypeBase::init();
}

class SortItem {
public:
	const Type* type;
	int predCount;
	void init(const Type* type)
	{
		this->type = type;
		this->predCount = type->getParents().size();
	}
};

void TypeManager::topologicalSortInst(const Type& root, vector<const Type*>& sorted)
{
	map<const Type*, SortItem> itemMap;
	vector<const Type*> zeroes;
	for (TypesIterator it = types.begin(); it != types.end(); it++) {
		if (it->second->isTypeOf(root)) {
			if (it->second == &root) {
				IOPC_DEBUGLOG(it->second->getName());
				zeroes.push_back(it->second);
			}
			itemMap[it->second].init(it->second);

		}
	}

	while (zeroes.size() != 0) {
		const Type* type = zeroes.back();
		IOPC_DEBUGLOG(type->getName());
		sorted.push_back(type);
		zeroes.pop_back();
		for (Type::ChildrenIterator it = type->getChildren().begin(); it != type->getChildren().end(); it++) {
			SortItem& item = itemMap[*it];
			item.predCount--;
			if (item.predCount == 0) {
				zeroes.push_back(*it);
			}
		}
	}
}

void TypeManager::parseTypeInst(string inStr, string& type, string& typeMeta, string& attr, string& attrMeta)
{
	vector<string> typeTokens;
	StringUtil::splitString(inStr, NAMESPACE_DELIMITER, typeTokens);
	string attrStr = typeTokens.back();
	typeTokens.pop_back();
	string typeStr = StringUtil::joinStrings(typeTokens, NAMESPACE_DELIMITER);
	parseMeta(typeStr, type, typeMeta);
	if (hasTypeInst(type)) {
		parseMeta(attrStr, attr, attrMeta);
		return;
	}
	parseMeta(inStr, type, typeMeta);
	attr = "";
	attrMeta = "";

	/*if () {
		string typeStr = inStr.substr(0, dotPos);
		string attrStr = inStr.substr(dotPos + 1);
		parseMeta(typeStr, type, typeMeta);
		parseMeta(attrStr, attr, attrMeta);
	} else {
		parseMeta(inStr, type, typeMeta);
	}*/
}

void TypeManager::parseMetaInst(string inStr, string& outStr, string& outMeta)
{
	int pos1 = inStr.find("[");
	int pos2 = inStr.find("]");
	if (pos1 <= 0 || pos2 <= 0 || pos1 > pos2) {
		outStr = inStr;
		outMeta = "";
		return;
	}
	outStr = inStr.substr(0, pos1);
	outMeta = inStr.substr(pos1 + 1, pos2 - pos1 - 1);
}


}

