/**
 * @file typeManager.h
 *
 * @ingroup iopcmeta
 *
 * @author Petr Cermak
 */

#ifndef TYPEMANAGER_H_
#define TYPEMANAGER_H_

#include "type.h"
#include <map>
#include <string>
#include <typeinfo>
#include <vector>
#include "iopccommon/lock.h"

namespace iopc {


/**
 * @brief Predicate for comparing two type_info pointers
 */
struct typeInfoCmp
{
	bool operator()(const std::type_info *t1, const std::type_info *t2) const {
		return t1->before(*t2) != 0;
	}
};

/**
 * @brief A singleton which maintains a list of @a Type Descriptions (instances of the @a Type subclasses).
 *
 * The class is thread safe for read-only access (i.e. no type registrations)
 * @ingroup iopcmeta
 */
class DLL_EXTERNAL TypeManager {
public:

	/**
	 * Keys of the map contains qualified type names, values contain the @a Type Descriptions
	 */
	typedef std::map<std::string, const Type*> TypesMap;
	typedef TypesMap::const_iterator TypesIterator;


	/**
	 * @brief Returns the singleton instance
	 */
	static TypeManager& getInstance();

	/**
	 * @brief Registers the specified type under a name provided by the @a Type Description
	 *
	 * Used by the @a TypeRegistrar instances generated by @c iopcsp
	 *
	 * @param type A pointer to the @a Type description
	 * @param ti typeid of the type being registered
	 */
	static inline void registerType(const Type* type, const std::type_info* ti) {
		getInstance().registerTypeInst(type, ti);
	}

	/**
	 * @brief Registers the specified type under a provided name
	 *
	 * @param type A pointer to the @a Type description
	 * @param typeName The name under which the type will be registered
	 * @param ti typeid of the type being registered
	 */
	static inline void registerType(const Type* type, const std::string& typeName, const std::type_info* ti) {
		getInstance().registerTypeInst(type, typeName, ti);
	}

	/**
	 * @brief Returns a @a Type object that represents a type specified by its name.
	 *
	 * The typeName parameter must be a fully qualified name.
	 * Example: "iopc::OidObject", "::TopLevelType"
	 *
	 * @param typeName A fully qualified type name.
	 * @return A type object that represents the type.
	 * @throws IopcNotFoundException if such type is not found
	 */
	static inline const Type& getType(const std::string& typeName) {
		return getInstance().getTypeInst(typeName);
	}

	/**
	 * @brief Checks whether the @a TypeManager contains a Type Description for the specified type
	 *
	 * @param typeName A fully qualified type name.
	 */
	static inline bool hasType(const std::string& typeName) {
		return getInstance().hasTypeInst(typeName);
	}

	/**
	 * @brief Returns a @a Type object that represents a type specified by its type_info object.
	 *
	 * This object is returned by the typeid operator.
	 *
	 * @param ti A type_info object
	 * @return A type object that represents the type.
	 * @throws IopcNotFoundException if such type is not found
	 */
	static inline const Type& getTypeByTypeInfo(const std::type_info* ti) {
		return getInstance().getTypeByTypeInfoInst(ti);
	}


	/**
	 * @brief Returns a map containing all registered types
	 *
	 * Its key contains qualified type names, values contain the @a Type Descriptions
	 */
	static const TypesMap& getTypesMap() {
		return getInstance().getTypesMapInst();
	}

	/**
	 * @brief Initializes type manager instances.
	 *
	 * Called by @a IopcMeta::init()
	 */
	static inline void init() {
		getInstance().initInst();
	}

	/**
	 * @brief Sorts all child types of a @a root type in a topological inheritance order.
	 * First go parents, then their children
	 *
	 * @param[in] root Root type of the inheritance tree to be sorted
	 * @param[out] sorted Topologically sorted list. Note that it does not include the root type.
	 */
	static inline void topologicalSort(const Type& root, std::vector<const Type*>& sorted) {
		getInstance().topologicalSortInst(root, sorted);
	}

	/**
	 * @brief Parses a type name, attribute name and metadata code from a string representation.
	 *
	 * Following formats are accepted:
	 * @code
	 * Type
	 * Type::attribute
	 * Type[metadata_code]::attribute
	 * Type::attribute[metadata_code]
	 * @endcode
	 *
	 * @c Type is a qualified type name. It may also be other arbitrary string
	 * in which case the attribute and attribute metadata are not parsed.
	 * If some of the compartments is missing, corresponding output parameter will be set to an empty string
	 *
	 * @param[in] inStr The string to be parsed
	 * @param[out] type Parsed type name
	 * @param[out] typeMeta Parsed type metadata code
	 * @param[out] attr Parsed attribute name
	 * @param[out] attrMeta Parsed attribute metadata code
	 *
	 */
	static inline void parseType(std::string inStr,  std::string& type,  std::string& typeMeta,  std::string& attr, std::string& attrMeta) {
		getInstance().parseTypeInst(inStr, type, typeMeta, attr, attrMeta);
	}

	void parseTypeInst(std::string inStr,  std::string& type,  std::string& typeMeta,  std::string& attr, std::string& attrMeta);
	void parseMetaInst(std::string inStr,  std::string& outStr,  std::string& outMeta);

	void registerTypeInst(const Type* type, const std::type_info* ti);
	void registerTypeInst(const Type* type, const std::string& typeName,const  std::type_info* ti);

	const Type& getTypeInst(const std::string& typeName);
	bool hasTypeInst(const std::string& typeName);
	const Type& getTypeByTypeInfoInst(const std::type_info* ti);

	const TypesMap& getTypesMapInst() const;

	void topologicalSortInst(const Type& root, std::vector<const Type*>& sorted);

	void initInst();

protected:
	static inline void parseMeta(std::string inStr,  std::string& outStr,  std::string& outMeta) {
		getInstance().parseMetaInst(inStr, outStr, outMeta);
	}

	TypeManager();
	~TypeManager();

private:
	std::map<std::string, const Type*> types;
	std::map<const std::type_info*, const Type*, struct typeInfoCmp> typeInfoToType;

	Lock lock;
};


}

#endif /*TYPEMANAGER_H_*/
